//#include "/usr/include/gsl/gsl_sf.h"

void GetVNLPoint(PointSetType * points, int index, VNLPointType *vnlp) //concatenates co-ordinates with feature values
{
  VectorType f;
  points->GetPointData(index,&f); 
  PointType p;
  points->GetPoint(index,&p);
  
  for (int k = 0; k < Dimension; k++)
    {
      (*vnlp)[k] = p[k];
    }
  for (int k = Dimension; k < GMMPointDimension; k++)
    {
      (*vnlp)[k] = f[k - Dimension];
    }
}

void SetVNLPoint(PointSetType * points, int index, VNLPointType *vnlp) //concatenates co-ordinates with feature values
{
  VectorType f;

  PointType p;
  
  for (int k = 0; k < Dimension; k++)
    {
       p[k] = (*vnlp)[k];
    }
  for (int k = Dimension; k < GMMPointDimension; k++)
    {
       f[k - Dimension] = (*vnlp)[k];
    }

  
  //points->GetPoints()->InsertElement(index,p);
  points->SetPoint(index,p);
  //points->GetPointData()->InsertElement(index,f);
  if ( GMMPointDimension > Dimension)
    points->SetPointData(index,f);  
  

}
void vnl2Xyz(VNLPointType *vnlp, XYZPointType *xyz)
{
  for(int i=0; i< Dimension; i++)
    (*xyz)[i]=(*vnlp)[i];
}
void xyz2Vnl(XYZPointType *xyz, VNLPointType *vnlp)
{
  for(int i=0; i< Dimension; i++)
    (*vnlp)[i] = (*xyz)[i];
  for(int i =Dimension; i< GMMPointDimension; i++)
    (*vnlp)[i] = 0.0;
}
void rotation2AffineMatrix(RotationMatrixType *rotMat,AffineMatrixType *affMat) 
{
  (*affMat).set_identity();
  for(int i=0;i<Dimension;i++)
    for(int j=0;j<Dimension;j++)
      (*affMat)[i][j] = (*rotMat)[i][j];
}
void UpdateRigidTransformations(PointSetVectorType * pointsetVector, VNLPointVectorType * modelPointsVector, 
			   AffineMatrixVectorType *affineMatrices, TranslationVectorType *translationVectors,
				PosteriorsMatrixVectorType * posteriorsVector)
{
  VNLPointVectorType::iterator modelPointsVectorIt;
  PointSetVectorType::iterator shapeIt;
  AffineMatrixVectorType::iterator affinesIt;
  TranslationVectorType ::iterator translationsIt;
  PosteriorsMatrixVectorType::iterator posteriorsIt;
 
  shapeIt = pointsetVector->begin();
  affinesIt = affineMatrices->begin();
  translationsIt = translationVectors->begin();
  posteriorsIt = posteriorsVector->begin();

  while( shapeIt != pointsetVector->end())
    {
      //Acquiring some info about current shape
      int Ns = (*shapeIt)->GetNumberOfPoints();
      AffineMatrixType Ak = (*affinesIt);
      PosteriorsMatrixType Ek = (*posteriorsIt);
      //computing barycenters
      XYZPointType sBar;
      XYZPointType mBar;
      sBar.fill(0.0);
      mBar.fill(0.0);
      for (int i = 0; i < Ns; i++)
	{
	  VNLPointType shapeVNLPoint;
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  XYZPointType xyzPoint;
	  vnl2Xyz(&shapeVNLPoint, &xyzPoint);
	  sBar += xyzPoint;
	  modelPointsVectorIt = modelPointsVector->begin();
	  while ( modelPointsVectorIt !=  modelPointsVector->end())
	    {
	      int j = modelPointsVectorIt - modelPointsVector->begin();
	      VNLPointType mj;
	      mj = (*modelPointsVectorIt);
	      XYZPointType mjX;
	      vnl2Xyz(&mj, &mjX);
	      mBar += Ek[i][j]*mjX;
	      ++modelPointsVectorIt;
	    }
	}
      sBar *= (1.0/(double)Ns);
      mBar *= (1.0/(double)Ns);
      VNLPointType vnlSBar;
      xyz2Vnl(&sBar,&vnlSBar);
      VNLPointType vnlMBar;
      xyz2Vnl(&mBar,&vnlMBar);
      (*translationsIt) = (vnlSBar - Ak*vnlMBar);
 
      //  std::cout << "vnlSBar" << (*tVec) << std::endl << std::endl;
      //now computing \Upsilon and \Gamma
      RotationMatrixType Upsilon;
      RotationMatrixType Gamma;
      Gamma.fill(0.0);
      Upsilon.fill(0.0);
      double denumS = 0.0;
      
      for (int i = 0; i < Ns; i++)
	{
	  VNLPointType shapeVNLPoint;
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  VNLPointType spi =shapeVNLPoint - (*translationsIt);
	  XYZPointType spiX;
	  vnl2Xyz( &spi, &spiX);
	  modelPointsVectorIt = modelPointsVector->begin();
	  while ( modelPointsVectorIt !=  modelPointsVector->end())
	    {
	      int j = modelPointsVectorIt - modelPointsVector->begin();
	      VNLPointType mj;
	      mj = (*modelPointsVectorIt);
	      XYZPointType mjX;
	      vnl2Xyz(&mj, &mjX);
	      Gamma += Ek[i][j]*outer_product(spiX, mjX);
	      Upsilon += Ek[i][j]*outer_product(mjX, mjX);
	      denumS +=  Ek[i][j]*mjX.squared_magnitude();
 	      ++modelPointsVectorIt;
	    }
	}
 
      
      RotationMatrixType U = vnl_svd<double>((vnl_matrix<double>)Gamma).U();
      RotationMatrixType VT = vnl_svd<double>((vnl_matrix<double>)Gamma).V().transpose();
      double  det = vnl_det(U)*vnl_det(VT);
      RotationMatrixType  S;
      S.set_identity();
      S[Dimension][Dimension] = det;
      RotationMatrixType Rot = U*S*VT;
      //std::cout << Rot << std::endl << std::endl;
      double numS = vnl_trace((vnl_matrix<double>)Gamma.transpose()*Rot);
      double s;
      if(denumS)
	{
	  s = numS/(denumS);
	  Rot *= s;
	}
      else
	{
	  Rot.set_identity();
	}

      AffineMatrixType Aff;
      rotation2AffineMatrix(&Rot, &Aff);
      
      /*
      RotationMatrixType Rot = Gamma*vnl_inverse(Upsilon);
      AffineMatrixType Aff;
      rotation2AffineMatrix(&Rot, &Aff);
      */
      
      (*affinesIt) = Aff;
      ++shapeIt;
      ++affinesIt;
      ++translationsIt;
      ++posteriorsIt;
    }
}

void UpdateSigma(PointSetVectorType * pointsetVector, VNLPointVectorType * modelPointsVector, 
		 AffineMatrixVectorType *affineMatrices, TranslationVectorType *translationVectors,
		 PosteriorsMatrixVectorType * posteriorsVector, double * sigV)//, double * sigmaf)
{

  PointSetVectorType::iterator shapeIt;
 
  shapeIt = pointsetVector->begin();

  PosteriorsMatrixVectorType::iterator posteriorsIt;
  posteriorsIt = posteriorsVector->begin();

  AffineMatrixVectorType::iterator affinesIt;
  affinesIt = affineMatrices->begin();

  TranslationVectorType ::iterator translationsIt;
  translationsIt = translationVectors->begin();

  VNLPointVectorType::iterator modelPointsVectorIt;
 

  //double S = 0.0;
  //double Sf= 0.0;
  int N = 0;
  double SV[GMMPointDimension];
  for(int i1=0; i1< GMMPointDimension; i1++)
    SV[i1]=0.0;
  while( shapeIt != pointsetVector->end() )
    {
      //Acquiring some info about current shape
      int Ns = (*shapeIt)->GetNumberOfPoints();
      AffineMatrixType Ak = (*affinesIt);
      TranslationType Tk = (*translationsIt);
      //std::cout << "Ak" <<  Ak << std::endl;
      //std::cout << "Tk" <<  Tk << std::endl;
      //Preparing model
      //int j = 0;
      modelPointsVectorIt = modelPointsVector->begin();
      while ( modelPointsVectorIt != modelPointsVector->end() )
	{
	  VNLPointType modelPoint = (*modelPointsVectorIt);
	  VNLPointType transformedModelPoint = Ak*modelPoint + Tk;
	  //std::cout << modelPoint << std::endl;
	  //std::cout << transformedModelPoint << std::endl;
	  //VNLPointType transformedModelPoint = modelPoint;
	  for (int i = 0; i < Ns; i++)
	    {
	      int j = modelPointsVectorIt -  modelPointsVector->begin();
	      VNLPointType shapeVNLPoint;
	      GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	      VNLPointType diff = shapeVNLPoint - transformedModelPoint;
	      double Ekij = (double)((*posteriorsIt)[i][j]);

	      for (int i1= 0; i1< GMMPointDimension; i1++)
		SV[i1] += diff[i1]*diff[i1]*Ekij;

	    }
	  ++modelPointsVectorIt;
	  //j += 1;
	}
      N = N + Ns;
      ++affinesIt;
      ++translationsIt;
      ++posteriorsIt;
      ++shapeIt;
    }
  
  if ( Dimension == 2 )
    {
      double V=0;
      for (int i1=0; i1<GMMPointDimension; i1++)
	V += SV[i1];
      for (int i1=0; i1<GMMPointDimension; i1++)
	{
	  sigV[i1] = vcl_sqrt(V/(double)GMMPointDimension/((double)(N-1)));
	  //if(sigmaV[i1] < 20.0) sigmaV[i1]= 1.0;
	}
    }
  else
    {
      double V=0;
      for (int i1=0; i1<Dimension; i1++)
	V += SV[i1];
      for (int i1=0; i1<Dimension; i1++)
	{
	  sigV[i1] = vcl_sqrt(V/(double)Dimension/(double)(N));
	  //std::cout <<  sigmaV[i1] << std::endl;
	}
      for (int i1=Dimension; i1<GMMPointDimension; i1++)
	sigV[i1] = vcl_sqrt(SV[i1]/(double)(N));
    }


  
} 
void  WriteTransformations(AffineMatrixVectorType *affineMatrices, TranslationVectorType *translationVectors,std::vector<char*> *meshNamesVector,char *name)
{
  int N = affineMatrices->end() - affineMatrices->begin();
  std::vector< std::ofstream *> filesVector;
  
  for(int i=0; i< N; i++)
    {

      std::string str= (*meshNamesVector)[i];
      unsigned found = str.find_last_of("/\\");
      str = str.substr(found+1);
      found = str.find_last_of(".");  
      
      char filename[1000];
      sprintf(filename,"%s_%s%s",str.substr(0,found).c_str(),name,".txt");
      //alignedShapesListFile << name << std::endl;
      std::ofstream * myfile= new std::ofstream;
      myfile->open(filename);
      filesVector.push_back(myfile);
    }

  AffineMatrixVectorType::iterator affinesIt;
  TranslationVectorType ::iterator translationsIt;
  std::vector< std::ofstream *>::iterator filesIt;

  filesIt = filesVector.begin();
  affinesIt = affineMatrices->begin();
  translationsIt = translationVectors->begin();
  while( affinesIt != affineMatrices->end())
    {
      *(*filesIt) << *translationsIt << std::endl;    
      *(*filesIt) << *affinesIt << std::endl;    
      ++filesIt;
      ++affinesIt;
      ++translationsIt;
    }
  for(int i=0; i< N; i++)
    {
      filesVector[i]->close();
      delete filesVector[i];
    }
  

}
void WriteCorrespondance(PointSetVectorType * pointsetVector, VNLPointVectorType * modelPointsVector, 
			   AffineMatrixVectorType *affineMatrices, TranslationVectorType *translationVectors,
			 PosteriorsMatrixVectorType * posteriorsVector, std::vector< std::ofstream *> *filesVector, char * virCorr)
{

  char myfilename[1000];
  sprintf(myfilename,"%s",virCorr);

  std::ofstream myfile;
  myfile.open(myfilename);

  int N = pointsetVector->end() - pointsetVector->begin();
  int Nm = modelPointsVector->end() - modelPointsVector->begin();

  //std::cout << "N: " << N << std::endl;
  //std::cout << " Nm: " << Nm << std::endl;

 

  VNLPointVectorType::iterator modelPointsVectorIt;
  PointSetVectorType::iterator shapeIt;
  AffineMatrixVectorType::iterator affinesIt;
  TranslationVectorType ::iterator translationsIt;
  PosteriorsMatrixVectorType::iterator posteriorsIt;
  std::vector< VNLPointVectorType >::iterator alignedShapeIt;

  AffineMatrixVectorType invAks;
  affinesIt = affineMatrices->begin();
  while( affinesIt != affineMatrices->end())
    {
      invAks.push_back(vnl_inverse(*affinesIt));
      ++affinesIt;
    }
  //now I change the iterator to refer to inverted matrices
  affinesIt = invAks.begin();
  std::vector< std::ofstream *>::iterator filesIt = filesVector->begin();

  std::vector< VNLPointVectorType > alignedPointsSetVector(N);


  VNLPointVectorType sumTiVector(Nm);
  std::vector<double> sumEVector(Nm);



  alignedShapeIt = alignedPointsSetVector.begin();
  shapeIt = pointsetVector->begin();
  translationsIt = translationVectors->begin();
  posteriorsIt = posteriorsVector->begin();

  while( shapeIt != pointsetVector->end())
    {
      int k = shapeIt - pointsetVector->begin();
      //Acquiring some info about current shape
      int Ns = (*shapeIt)->GetNumberOfPoints();
      //AffineMatrixType Ak_1 = (*affinesIt);
      AffineMatrixType Ak_1 = (*affinesIt);
      PosteriorsMatrixType Ek = (*posteriorsIt);
      TranslationType Tk = (*translationsIt);

      for(int i=0; i< Nm; i++)
	{
	  sumTiVector[i].fill(0);
	  sumEVector[i] = 0.0;

	}

      for (int i = 0; i < Ns; i++)
	{
	  VNLPointType shapeVNLPoint;
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  modelPointsVectorIt = modelPointsVector->begin();
	  while ( modelPointsVectorIt !=  modelPointsVector->end())
	    {
	      int j = modelPointsVectorIt - modelPointsVector->begin();
	      sumTiVector[j] += (Ek[i][j]+MYEPS)*(Ak_1*(shapeVNLPoint - Tk));
	      sumEVector[j] += (Ek[i][j]+MYEPS);
	      ++modelPointsVectorIt;
	    }
	} 	    
      modelPointsVectorIt = modelPointsVector->begin();
      while ( modelPointsVectorIt !=  modelPointsVector->end())
	{
	  int j = modelPointsVectorIt - modelPointsVector->begin();
	  VNLPointType invShapeVNLPoint;
	  invShapeVNLPoint = (double)(1.0/(sumEVector[j]))*sumTiVector[j];
	  //invShapeVNLPoint = (double)(1.0/sumEVector[j])*sumTiVector[j];
	  (*alignedShapeIt).push_back(invShapeVNLPoint);
	  ++modelPointsVectorIt;
	}
      
      ++posteriorsIt;
      ++affinesIt;
      ++alignedShapeIt;
      ++translationsIt;
      ++shapeIt;
    }

  modelPointsVectorIt = modelPointsVector->begin();
  while ( modelPointsVectorIt !=  modelPointsVector->end())
    {
      int j = modelPointsVectorIt - modelPointsVector->begin();
      myfile << (*modelPointsVectorIt) << " ";
      for (int k=0; k<N; k++)
	{
	  myfile << alignedPointsSetVector[k][j] << " ";
	  (*(*filesVector)[k]) << alignedPointsSetVector[k][j] << std::endl;
	}
      myfile << std::endl;
      ++modelPointsVectorIt;
    }
  myfile.close();
  //std::cout << "here4" << std::endl;
  for(int i=0; i< N; i++)
    {
      (*filesVector)[i]->close();
      //delete (*filesVector)[i];
    }

 


}

void TransformShapes(PointSetVectorType * pointSetVector, AffineMatrixVectorType *affineMatrices, 
		       TranslationVectorType *translationVectors, std::vector< std::ofstream *> *filesVector)
{
 
  PointSetVectorType::iterator shapeIt=pointSetVector->begin();
  AffineMatrixVectorType::iterator affinesIt;
  TranslationVectorType ::iterator translationsIt = translationVectors->begin();

  AffineMatrixVectorType invAks;
  affinesIt = affineMatrices->begin();
  while( affinesIt != affineMatrices->end())
    {
      invAks.push_back(vnl_inverse(*affinesIt));
      ++affinesIt;
    }
  //now I change the iterator to refer to inverted matrices
  affinesIt = invAks.begin();
 

  while( shapeIt != pointSetVector->end())
    {
     int k = shapeIt - pointSetVector->begin();
     //Acquiring some info about current shape
     int Ns = (*shapeIt)->GetNumberOfPoints();
     AffineMatrixType Ak_1 = (*affinesIt);
     TranslationType Tk = (*translationsIt);
     for (int i = 0; i < Ns; i++)
       {
	 VNLPointType shapeVNLPoint;
	 GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	 (*(*filesVector)[k]) << Ak_1*(shapeVNLPoint - Tk) << std::endl;
       }
      ++affinesIt;
      ++translationsIt;
      ++shapeIt;
    }
	 
  int N = affineMatrices->end() - affineMatrices->begin();
  for(int i=0; i< N; i++)
    {
      (*filesVector)[i]->close();
      //delete filesVector[i];
    }



}

void UpdateTransformations(PointSetVectorType * pointsetVector, VNLPointVectorType * modelPointsVector, 
			   AffineMatrixVectorType *affineMatrices, TranslationVectorType *translationVectors,
			   PosteriorsMatrixVectorType * posteriorsVector)
{
  VNLPointVectorType::iterator modelPointsVectorIt;
  PointSetVectorType::iterator shapeIt;
  AffineMatrixVectorType::iterator affinesIt;
  TranslationVectorType ::iterator translationsIt;
  PosteriorsMatrixVectorType::iterator posteriorsIt;


  shapeIt = pointsetVector->begin();
  affinesIt = affineMatrices->begin();
  translationsIt = translationVectors->begin();
  posteriorsIt = posteriorsVector->begin();

  while( shapeIt != pointsetVector->end())
    {
      //Acquiring some info about current shape
      int Ns = (*shapeIt)->GetNumberOfPoints();
      AffineMatrixType Ak = (*affinesIt);
      PosteriorsMatrixType Ek = (*posteriorsIt);
      //computing barycenters
      VNLPointType sBar;
      VNLPointType mBar;
      sBar.fill(0.0);
      mBar.fill(0.0);
      for (int i = 0; i < Ns; i++)
	{
	  VNLPointType shapeVNLPoint;
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  sBar += shapeVNLPoint;
	  modelPointsVectorIt = modelPointsVector->begin();
	  while ( modelPointsVectorIt !=  modelPointsVector->end())
	    {
	      int j = modelPointsVectorIt - modelPointsVector->begin();
	      VNLPointType mj;
	      mj = (*modelPointsVectorIt);
	      mBar += Ek[i][j]*mj;
	      ++modelPointsVectorIt;
	    }
	}
      sBar *= (1.0/(double)Ns);
      mBar *= (1.0/(double)Ns);
      (*translationsIt) = (sBar - Ak*mBar);
      
      //now computing \Upsilon and \Gamma
      AffineMatrixType Upsilon;
      AffineMatrixType Gamma;
      Gamma.fill(0.0);
      Upsilon.fill(0.0);
      
      for (int i = 0; i < Ns; i++)
	{
	  VNLPointType shapeVNLPoint;
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  //VNLPointType spi =shapeVNLPoint - sBar;
	  VNLPointType spi =shapeVNLPoint - (*translationsIt);
	  modelPointsVectorIt = modelPointsVector->begin();
	  while ( modelPointsVectorIt !=  modelPointsVector->end())
	    {
	      int j = modelPointsVectorIt - modelPointsVector->begin();
	      VNLPointType mj;
	      mj = (*modelPointsVectorIt);
	      //VNLPointType mpj = mj - mBar;
	      VNLPointType mpj = mj;
	      AffineMatrixType  sm,mm;
	      Gamma += Ek[i][j]*outer_product(spi, mpj); //correct
	      Upsilon += Ek[i][j]*outer_product(mpj, mpj);
 	      ++modelPointsVectorIt;
	    }
	}
 

      vnl_matrix_fixed<double, GMMPointDimension, GMMPointDimension> U = vnl_svd<double>((vnl_matrix<double>)Gamma).U();
      vnl_matrix_fixed<double, GMMPointDimension, GMMPointDimension> VT = vnl_svd<double>((vnl_matrix<double>)Gamma).V().transpose();
      double  det = vnl_det(U)*vnl_det(VT);
      vnl_matrix_fixed<double, GMMPointDimension, GMMPointDimension>  I;
      I.set_identity();
    
      if (det < 0.0) //det==-1
	I[GMMPointDimension][GMMPointDimension] = -1.0;

      *affinesIt = U*I*VT;

      

      //*affinesIt = vnl_svd<double>((vnl_matrix<double>)Gamma).U();
      //*affinesIt *= vnl_svd<double>((vnl_matrix<double>)Gamma).V().transpose();
      //double trGamma =  vnl_trace((vnl_matrix<double>)Gamma);
      //double trUpsilon =  vnl_trace((vnl_matrix<double>)Upsilon);
      //*affinesIt *= (trGamma/(trUpsilon+MYEPS)); //most stupid mistake i have ever made in my programming!!
 
     /*
      std::cout << "shape: " <<  shapeIt - pointsetVector->begin() << std::endl;
      std::cout << "translation: " << std::endl << *translationsIt << std::endl;
      std::cout << "affine matrix: " << std::endl << *affinesIt << std::endl << std::endl;
      */

      ++shapeIt;
      ++affinesIt;
      ++translationsIt;
      ++posteriorsIt;
    }
}




void UpdatePosteriorsVector(PointSetVectorType * pointsetVector, VNLPointVectorType * modelPointsVector, 
			    AffineMatrixVectorType *affineMatrices, TranslationVectorType *translationVectors,
			    PosteriorsMatrixVectorType * posteriorsVector, PriorsVectorType * priorsVector, double *sigmaV)//double * sigma, double *sigmaf)
{

  PointSetVectorType::iterator shapeIt;
  shapeIt = pointsetVector->begin();

  PosteriorsMatrixVectorType::iterator posteriorsIt;
  posteriorsIt = posteriorsVector->begin();

  AffineMatrixVectorType::iterator affinesIt;
  affinesIt = affineMatrices->begin();

  TranslationVectorType ::iterator translationsIt;
  translationsIt = translationVectors->begin();

  VNLPointVectorType::iterator modelPointsVectorIt;
  PriorsVectorType::iterator priorsVectorIt;

  //double sig = (*sigma);
  //double sigf = (*sigmaf);
  double *sig = sigmaV;

  while( shapeIt != pointsetVector->end() )
    {
      //Acquiring some info about current shape
      int Ns = (*shapeIt)->GetNumberOfPoints();
      AffineMatrixType Ak = (*affinesIt);
      TranslationType Tk = (*translationsIt);
      for (int i = 0; i < Ns; i++)
	{
	  //PointType shapePoint = (*shapeIt)->GetPoint(i);
	  VNLPointType shapeVNLPoint;
	  //GetVNLPoint(&shapePoint, &shapeVNLPoint);
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  double denum = 0.0;
	  
	  //compute denuminator
	  modelPointsVectorIt = modelPointsVector->begin();
	  priorsVectorIt = priorsVector->begin();
	  while ( modelPointsVectorIt != modelPointsVector->end() )
	    {
	      int j = modelPointsVectorIt -  modelPointsVector->begin();
	      VNLPointType modelPoint = (*modelPointsVectorIt);
	      double prior_j = (*priorsVectorIt);
	      VNLPointType transformedModelPoint = Ak*modelPoint + Tk;
	      VNLPointType diff = shapeVNLPoint - transformedModelPoint;
	      //denum += exp(diff.squared_magnitude()/(-2.0*sig*sig))*prior_j;
	      /*double d = 0.0;
	      double df = 0.0;
	      for(int i1=0; i1<Dimension; i1++)
		d +=diff[i1]*diff[i1];
	      for(int i1=Dimension; i1< GMMPointDimension; i1++)
		df +=diff[i1]*diff[i1];
		denum += exp(d/(-2.0*sig*sig))*exp(df/(-2.0*sigf*sigf))*prior_j;*/
	      double d[GMMPointDimension];
	      for(int i1=0; i1<GMMPointDimension; i1++)
		d[i1] = diff[i1]*diff[i1];
	      double p =1.0;
	      for(int i1=0; i1<GMMPointDimension; i1++)
		p *= exp(d[i1]/(-2.0*sig[i1]*sig[i1]));
	      denum += (p*prior_j);
	      
	      ++modelPointsVectorIt;
	      ++priorsVectorIt;

	    }//while

	  if (!denum)
	    {
	      //computing numenator
	      modelPointsVectorIt = modelPointsVector->begin();
	      while ( modelPointsVectorIt != modelPointsVector->end() )
		{
		  int j = modelPointsVectorIt -  modelPointsVector->begin();
		  (*posteriorsIt)[i][j] = 1.0/(double)(modelPointsVector->end() -  modelPointsVector->begin());
		  ++modelPointsVectorIt;
		}//while
	    }
	  else
	    {
	  
	      //computing numenator
	      modelPointsVectorIt = modelPointsVector->begin();
	      priorsVectorIt = priorsVector->begin();
	      while ( modelPointsVectorIt != modelPointsVector->end() )
		{
		  int j = modelPointsVectorIt -  modelPointsVector->begin();
		  VNLPointType modelPoint = (*modelPointsVectorIt);
		  double prior_j = (*priorsVectorIt);
		  VNLPointType transformedModelPoint = Ak*modelPoint + Tk;
		  VNLPointType diff = shapeVNLPoint - transformedModelPoint;
		  double d[GMMPointDimension];
		  for(int i1=0; i1<GMMPointDimension; i1++)
		    d[i1] = diff[i1]*diff[i1];
		  double p =1.0;
		  for(int i1=0; i1<GMMPointDimension; i1++)
		    p *= exp(d[i1]/(-2.0*sig[i1]*sig[i1]));
		  double num = (p*prior_j);
		  
		  (*posteriorsIt)[i][j] = num/(denum);
		  ++modelPointsVectorIt;
		  ++priorsVectorIt;
		}//while
	    }//else
	}//for
      ++affinesIt;
      ++translationsIt;
      ++posteriorsIt;
      ++shapeIt;

    }//while

}

int PruneModel(VNLPointVectorType * modelPointsVector, PriorsVectorType * priorsVector, PriorsVectorType * uncPriorsVector,PosteriorsMatrixVectorType * posteriorsVector)
{
  VNLPointVectorType::iterator modelPointsVectorIt;
  PriorsVectorType::iterator priorsVectorIt;
  PriorsVectorType::iterator uncPriorsVectorIt;

  priorsVectorIt = priorsVector->begin();
  modelPointsVectorIt = modelPointsVector->begin();
  uncPriorsVectorIt = uncPriorsVector->begin();
  int M = 0;
  while( priorsVectorIt != priorsVector->end())
    {
      if ( *priorsVectorIt != 0 )  M++;
      ++priorsVectorIt;
    }

  int oM =priorsVector->end() - priorsVector->begin();
  //printf("oM: %d M: %d\n",oM,M);

  int nb_train = posteriorsVector->size();
  PosteriorsMatrixVectorType posteriorsBufferVector;
  posteriorsBufferVector.resize(nb_train);
  // int oM = (*posteriorsVector)[0].columns();
  for ( int k = 0; k < nb_train; k++)
    {
      posteriorsBufferVector[k].set_size((*posteriorsVector)[k].rows(), (*posteriorsVector)[k].columns() );
      posteriorsBufferVector[k].copy_in( (*posteriorsVector)[k].data_block());
      /*
      int Ns = (*posteriorsVector)[k].rows();
      PosteriorsMatrixType post(Ns, oM);
      for(int i=0; i< Ns; i++)
	for(int j=0; j<oM; j++)
	  post[i][j]=(*posteriorsVector)[k][i][j];
      posteriorsBufferVector.push_back(post);
      */
    }
  //printf("oM: %d M: %d\n",oM,M);
  //now resizing the posteriorsVector
  for ( int k = 0; k < nb_train; k++)
    {
      int Ns = (*posteriorsVector)[k].rows();
      (*posteriorsVector)[k].set_size(Ns,M);
      int c = 0;
      for(int j=0; j < oM; j++)
	if( (*priorsVector)[j] != 0.0)
	    {
	      (*posteriorsVector)[k].set_column(c, posteriorsBufferVector[k].get_column(j));
	      c++;
	    }
    }  
  int sum=0;
  priorsVectorIt = priorsVector->begin();
  modelPointsVectorIt = modelPointsVector->begin();
  uncPriorsVectorIt = uncPriorsVector->begin();
  //printf("oM: %d M: %d\n",oM,M);


  while( priorsVectorIt != priorsVector->end())
    {
      if ( *priorsVectorIt == 0 ) 
	{
	  priorsVector->erase(priorsVectorIt);
	  modelPointsVector->erase(modelPointsVectorIt);
	  uncPriorsVector->erase(uncPriorsVectorIt);
	  sum++;
	}
      else
	{
	  ++modelPointsVectorIt;
	  ++priorsVectorIt;
	  ++uncPriorsVectorIt;
	}
    }
  posteriorsBufferVector.resize(0);

  return sum;
}

void  UpdateUnconstrainedPriorsVector(PosteriorsMatrixVectorType * posteriorsVector, PriorsVectorType * uncPriorsVector, const double *alpha)
{
  PosteriorsMatrixVectorType::iterator posteriorsIt;
  posteriorsIt = posteriorsVector->begin();

  PriorsVectorType::iterator uncPriorsVectorIt;
  uncPriorsVectorIt = uncPriorsVector->begin();

  double a = (*alpha);
  
  int Nm = uncPriorsVector->end() - uncPriorsVector->begin();
 
  int sum = 0;
  while( uncPriorsVectorIt != uncPriorsVector->end() )
    {
      int j = uncPriorsVectorIt - uncPriorsVector->begin();
      double sE = 0.0;
      int NxNs = 0;
      posteriorsIt = posteriorsVector->begin();
      while( posteriorsIt != posteriorsVector->end() )
	{
	  int Ns = (*posteriorsIt).rows();
	  NxNs += Ns;
	  for (int i = 0; i < Ns; i++)
	    {
	      sE += (*posteriorsIt)[i][j];
	    }
	  ++posteriorsIt;
	}
      //  std::cout << "sE = " << sE << std::endl;
      (*uncPriorsVectorIt) = (sE + NxNs*(a - 1.0))/(Nm*NxNs*(a - 1.0) + NxNs);
      //sum += (*uncPriorsVectorIt);
      if ( (*uncPriorsVectorIt) < 0.0 )
	{
	  //(*uncPriorsVectorIt) = 0.0;
	  sum += 1;
	}
      ++uncPriorsVectorIt;
    }
}
double MeasurePriorsEntropy(PriorsVectorType * priorsVector)
{
  int Nm = (*priorsVector).size();
  double trInv = 0.0;
  for (int j=0; j<Nm; j++)
    trInv += 1.0/(*priorsVector)[j];
  double S = 0.0;
  double Si = 0.0;
  for (int j=0; j<Nm; j++)
    {
      double pi = (*priorsVector)[j];
      double pi_1 = 1.0/pi/trInv;
      S += pi*log(pi);
      Si += pi_1*log(pi_1);
    }
  //std::cout << "S+Si: " << S+Si << std::endl;

  return ( S + Si + 2.0*log(Nm));
}
double MeasureEigenValuesEntropy(PointSetVectorType *pointSetVector, VNLPointVectorType * modelPointsVector,
			AffineMatrixVectorType *affineMatrices, TranslationVectorType *translationVectors,
			PosteriorsMatrixVectorType * posteriorsVector,int r)
{
  //char myfilename[1000]="tmpCorr.txt";
  char myfilename[1000];
  sprintf(myfilename,"tmpCorr_%02d.txt",r+1);
  std::ofstream myfile;
  myfile.open(myfilename);

  int N = pointSetVector->end() - pointSetVector->begin();
  int Nm = modelPointsVector->end() - modelPointsVector->begin();

  VNLPointVectorType::iterator modelPointsVectorIt;
  PointSetVectorType::iterator shapeIt;
  AffineMatrixVectorType::iterator affinesIt;
  TranslationVectorType ::iterator translationsIt;
  PosteriorsMatrixVectorType::iterator posteriorsIt;
  std::vector< VNLPointVectorType >::iterator alignedShapeIt;

  AffineMatrixVectorType invAks;
  affinesIt = affineMatrices->begin();
  while( affinesIt != affineMatrices->end())
    {
      invAks.push_back(vnl_inverse(*affinesIt));
      ++affinesIt;
    }
  //now I change the iterator to refer to inverted matrices
  affinesIt = invAks.begin();


  std::vector< VNLPointVectorType > alignedPointsSetVector(N);


  VNLPointVectorType sumTiVector(Nm);
  std::vector<double> sumEVector(Nm);



  alignedShapeIt = alignedPointsSetVector.begin();
  shapeIt = pointSetVector->begin();
  translationsIt = translationVectors->begin();
  posteriorsIt = posteriorsVector->begin();

  while( shapeIt != pointSetVector->end())
    {
      int k = shapeIt - pointSetVector->begin();
      //Acquiring some info about current shape
      int Ns = (*shapeIt)->GetNumberOfPoints();
      //AffineMatrixType Ak_1 = (*affinesIt);
      AffineMatrixType Ak_1 = (*affinesIt);
      PosteriorsMatrixType Ek = (*posteriorsIt);
      TranslationType Tk = (*translationsIt);

      for(int i=0; i< Nm; i++)
	{
	  sumTiVector[i].fill(0);
	  sumEVector[i] = 0.0;

	}

      for (int i = 0; i < Ns; i++)
	{
	  VNLPointType shapeVNLPoint;
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  modelPointsVectorIt = modelPointsVector->begin();
	  while ( modelPointsVectorIt !=  modelPointsVector->end())
	    {
	      int j = modelPointsVectorIt - modelPointsVector->begin();
	      sumTiVector[j] += (Ek[i][j]+MYEPS)*(Ak_1*(shapeVNLPoint - Tk));
	      sumEVector[j] += (Ek[i][j]+MYEPS);
	      ++modelPointsVectorIt;
	    }
	} 	    
      modelPointsVectorIt = modelPointsVector->begin();
      while ( modelPointsVectorIt !=  modelPointsVector->end())
	{
	  int j = modelPointsVectorIt - modelPointsVector->begin();
	  VNLPointType invShapeVNLPoint;
	  invShapeVNLPoint = (double)(1.0/(sumEVector[j]))*sumTiVector[j];
	  //invShapeVNLPoint = (double)(1.0/sumEVector[j])*sumTiVector[j];
	  (*alignedShapeIt).push_back(invShapeVNLPoint);
	  ++modelPointsVectorIt;
	}
      
      ++posteriorsIt;
      ++affinesIt;
      ++alignedShapeIt;
      ++translationsIt;
      ++shapeIt;
    }

  modelPointsVectorIt = modelPointsVector->begin();
  while ( modelPointsVectorIt !=  modelPointsVector->end())
    {
      int j = modelPointsVectorIt - modelPointsVector->begin();
      myfile << (*modelPointsVectorIt) << " ";
      for (int k=0; k<N; k++)
	{
	  myfile << alignedPointsSetVector[k][j] << " ";
	}
      myfile << std::endl;
      ++modelPointsVectorIt;
    }
  myfile.close();

  //now I have the temporary correspondance file
  
  char command[1000];
  char tmpKLName[1000];
  sprintf(tmpKLName,"tmpKL_%02d.txt", r+1);
  sprintf(command,"ComputePCA %s tmpPCA_%02d.txt %s", myfilename, r+1, tmpKLName);
  int i=system(command);
  
  //open the file
  std::ifstream klFile;
  klFile.open(tmpKLName);
  double kl=0;
  klFile >> kl;
 
  

  return kl;
}

void UpdateModelPointsVector(PointSetVectorType * pointSetVector, VNLPointVectorType * modelPointsVector, 
			     AffineMatrixVectorType *affineMatrices, TranslationVectorType *translationVectors,
			     PosteriorsMatrixVectorType * posteriorsVector)
//			     PosteriorsMatrixVectorType * posteriorsVector, PriorsVectorType * priorsVector)
{
  PointSetVectorType::iterator shapeIt;

  PosteriorsMatrixVectorType::iterator posteriorsIt;


  AffineMatrixVectorType::iterator affinesIt;

  TranslationVectorType ::iterator translationsIt;
 
  VNLPointVectorType::iterator modelPointsVectorIt;
  //PriorsVectorType::iterator priorsVectorIt;
  
  modelPointsVectorIt = modelPointsVector->begin();
  //priorsVectorIt = priorsVector->begin();
  while( modelPointsVectorIt != modelPointsVector->end())
    {
      //  if ( *priorsVectorIt != 0 ) // we need to update m_j
      //	{
	  int j =  modelPointsVectorIt - modelPointsVector->begin();
	  shapeIt = pointSetVector->begin();
	  posteriorsIt = posteriorsVector->begin();	      
	  affinesIt = affineMatrices->begin();
	  translationsIt = translationVectors->begin();
	  AffineMatrixType sumAkAk;
	  sumAkAk.fill(0.0);
	  VNLPointType sumAkTk;
	  sumAkTk.fill(0.0);
	  while( shapeIt != pointSetVector->end() )
	    {

	      AffineMatrixType Ak = (*affinesIt);
	      TranslationType Tk = (*translationsIt);

	      int Ns = (*posteriorsIt).rows();
	      for (int i = 0; i < Ns; i++)
		{

		  VNLPointType shapeVNLPoint;

		  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
		  sumAkTk += (*posteriorsIt)[i][j]*(Ak.transpose()*(shapeVNLPoint - Tk ));
		  sumAkAk += (*posteriorsIt)[i][j]*Ak.transpose()*Ak;
		}
	      ++posteriorsIt;
	      ++shapeIt;
	      ++affinesIt;
	      ++translationsIt;
	    }
	  (*modelPointsVectorIt) = vnl_inverse(sumAkAk)*sumAkTk;
	  //	}
	  //++priorsVectorIt;
      ++modelPointsVectorIt;
     
    }

}

void UpdateMixtureCoeffs( PosteriorsMatrixVectorType * posteriorsVector, PriorsVectorType * priorsVector)
{
    PosteriorsMatrixVectorType::iterator posteriorsIt;
    PriorsVectorType::iterator priorsIt;
    double NP = 0.0;
    //std::cout << "Code is here" << std::endl;
    int Nm = priorsVector->end() - priorsVector->begin();
    double ssPP[Nm];
     for (int j=0;j<Nm;j++) 
     {
       ssPP[j]=0.0;
     }

   
     priorsIt = priorsVector->begin();
     while(priorsIt != priorsVector->end()) // Iterates over training point sets
       {
	 int j = priorsIt - priorsVector->begin();
	 posteriorsIt = posteriorsVector->begin();

	 while(posteriorsIt != posteriorsVector->end())
	   {
	     int Ns = (*posteriorsIt).rows();
	     for (int i=0; i<Ns; i++)
	       {
		 ssPP[j] += ((*posteriorsIt)[i][j]);
		 NP += ((*posteriorsIt)[i][j]);
	       }
	     ++posteriorsIt;
	   }	
	  ++priorsIt;
       }
     
     //std::cout << "Sum of Mixture coefficients: " << NP << std::endl;

     priorsIt = priorsVector->begin();
     while(priorsIt != priorsVector->end())
       {
	 int j = priorsIt - priorsVector->begin();
     	 (*priorsIt) = ((ssPP[j])/NP);
	 //std::cout << "Mixture coefficient: " << (*priorsIt) << std::endl;
	 ++priorsIt;
       }
    

}

double  ComputeBIC(PointSetVectorType *pointsetVector, VNLPointVectorType * modelPointsVector,
	       AffineMatrixVectorType *affineMatrices, TranslationVectorType *translationVectors,
	       PriorsVectorType * priorsVector, double *sigmaV, double *bic, double *alpha)
{

  PointSetVectorType::iterator shapeIt;
  shapeIt = pointsetVector->begin();


  AffineMatrixVectorType::iterator affinesIt;
  affinesIt = affineMatrices->begin();

  TranslationVectorType ::iterator translationsIt;
  translationsIt = translationVectors->begin();

  VNLPointVectorType::iterator modelPointsVectorIt;
  PriorsVectorType::iterator priorsVectorIt;

  double *sig = sigmaV;
  int N = 0;

  while( shapeIt != pointsetVector->end() )
    {
      //Acquiring some info about current shape
      int Ns = (*shapeIt)->GetNumberOfPoints();
      N += Ns;
      AffineMatrixType Ak = (*affinesIt);
      TranslationType Tk = (*translationsIt);
      double sumLn = 0.0;
      for (int i = 0; i < Ns; i++)
	{
	  VNLPointType shapeVNLPoint;
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  double sumP=0.0;
	  modelPointsVectorIt = modelPointsVector->begin();
	  priorsVectorIt = priorsVector->begin();
	  while ( modelPointsVectorIt != modelPointsVector->end() )
	    {

	      int j = modelPointsVectorIt -  modelPointsVector->begin();
	      VNLPointType modelPoint = (*modelPointsVectorIt);
	      double prior_j = (*priorsVectorIt);
	      VNLPointType transformedModelPoint = Ak*modelPoint + Tk;
	      VNLPointType diff = shapeVNLPoint - transformedModelPoint;
	      double d[GMMPointDimension];
	      for(int i1=0; i1<GMMPointDimension; i1++)
		d[i1] = diff[i1]*diff[i1];
	      double p = 1.0;
	      for(int i1=0; i1<GMMPointDimension; i1++)
		{
		  p *= exp(d[i1]/(-2.0*sig[i1]*sig[i1]));
		  p /= (sqrt(2.0*3.14195)*sig[i1]);
		}
	      p *= prior_j;
	      sumP += p;
	      ++modelPointsVectorIt;
	    }//while model points
	  sumLn += log(sumP);
	}//for shape points
      int k = shapeIt - pointsetVector->begin();
      *bic += sumLn;
      ++affinesIt;
      ++translationsIt;
      ++shapeIt;
    }//while

  //determining number of free parameters
  double Nm = priorsVector->end() - priorsVector->begin();
  double K = pointsetVector->end() - pointsetVector->begin();

  //n = pi + mu + sigma + T + Rot + s
  double nu =0.0;
  if (GMMPointDimension == 4 )
    nu += Nm - 1.0 + Nm*4.0 + 2.0 + K*4.0 + 6.0*K + K;
  // else if (GMMPointDimension == 3 )
  //  nu += Nm - 1.0 + Nm*GMMPointDimension + 1.0 + K*GMMPointDimension + 3.0*K + K;

  //*bic = 2.0*(*bic) - nu*log((double)N);
  //return (*bic);
  double aic = 2.0*(*bic) - 2.0*nu;
  return (aic);
    
}
