/*
Author: Ali Gooya <ali dot gooya at gmail dot com>
*/
#include "stdafx.h"

#include "TypeDefinitions.h"
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <string>
#include <linalg.h>
using namespace alglib;

void print_help()
{
      std::cerr << std::endl;
      std::cerr << " Computes eignvalues and vectors from a file containing aligend samples." << std::endl;
      std::cerr << " The input file format is: x1 | x2 | ..." << std::endl;
      std::cerr << " Each row of xn includes: x y z f1 f2 .." << std::endl;
      std::cerr << " Usage: computePCA " << std::endl; 
      std::cerr << " -i inputAlignedXs.txt " << std::endl;
      std::cerr << " -o outputPCA.txt " << std::endl;
      std::cerr << std::endl;
}

int main( int argc, char *argv[] )
{

  int c = 0;
  int rOpts = 0;
  char inputFileName[1000];
  char outputFileName[1000];
  while (( c=getopt (argc, argv, "i:o:")) != -1)
    switch (c)
      {
      case 'i':
	sprintf(inputFileName,"%s",optarg);
	++rOpts;
	break;
      case 'o':
	sprintf(outputFileName,"%s",optarg);
	++rOpts;
	break;
      case '?':
	fprintf (stderr, "Unknown option %s.\n", optarg);
	return 1;
      default:
	abort ();
      }
  if (rOpts != 2)
    {
      print_help();
      return 1;
    }
	
  //open the input file and read it;

  VNLPointVectorType points;
  std::vector<VNLPointVectorType> pointSets;
  std::ifstream inpfile;
  inpfile.open(inputFileName);
  bool firstline = true;
  int Ns;
  std::string   line;
  while(std::getline(inpfile, line))
    {
      std::stringstream linestream(line);
      VNLPointType vnlp;
      while(linestream >> vnlp)
	points.push_back(vnlp);
      if(firstline)
	{
	  Ns =  points.end() - points.begin();
	  //  std::cout<< points << std::endl;
	  for ( int i = 0; i < Ns; i++)
	    {
	      VNLPointVectorType shape;
	      shape.push_back(points[i]);
	      pointSets.push_back(shape);
	    }
 	  firstline = false;
	}
      else
	{
	  for ( int i = 0; i < Ns; i++)
	    {
	      pointSets[i].push_back(points[i]);
	    }
	}
      points.resize(0);
    }
  
  //for(int i =0; i< pointSets[0].end() - pointSets[0].begin(); i++)
  //  std::cout << pointSets[0][i] << std::endl;
  int Nm = pointSets[0].end() - pointSets[0].begin();

  //computing xbar
  VNLPointVectorType xbar(Nm);
  for ( int i=0; i< Dimension; i++)
    for ( int j=0; j< Nm; j++)
      xbar[j][i] = 0.0;
  for ( int i=0; i< Ns; i++)
    for ( int j=0; j< Nm; j++)
	xbar[j] += pointSets[i][j];

  //std::cout << xbar[0]*(1.0/((double)Ns)) << std::endl; 
  //now taking xbar off from xs
  for ( int i=0; i< Ns; i++)
    for ( int j=0; j< Nm; j++)
	{
	  pointSets[i][j] -=  xbar[j]/(double)Ns;
	}  

  //double** ary = new double*[Nm*Dimension];

 
  //for(int i = 0; i < Nm*Dimension; ++i)
  //  ary[i] = new double[Ns];

  real_2d_array ARY;
  ARY.setlength(Nm*Dimension,Ns);
  for ( int i=0; i< Ns; i++)
    for ( int j=0; j< Nm; j++)
      for ( int k =0; k<Dimension; k++)
	ARY[j*Dimension+k][i] = pointSets[i][j][k];
  
 
  //double *w = new double[Nm*Dimension];
  real_1d_array W;
  W.setlength(Ns);

  real_2d_array U;
  U.setlength(Nm*Dimension,Nm*Dimension);
  real_2d_array VT;
  VT.setlength(Ns,Ns);

  rmatrixsvd(ARY,Nm*Dimension,Ns,1,1,2,W,U,VT);
  //for( int i=0; i< Ns; i++)
  //  printf(" %f ",W[i]);
  //printf("\n");
  
  //now writing results to pca model file
  std::ofstream mypcafile;
  mypcafile.open(outputFileName);
  
  mypcafile << Nm + 1 << std::endl;
  
  for( int i=0; i< Nm; i++)
    {
      mypcafile << xbar[i]*(1.0/((double)Ns)) << "  ";
    }
  mypcafile << std::endl;

  /* 
  double S1 = 0.0;
  double S2 = 0.0;
  int D = Nm;
  int N = Ns-1;
   
  double trace = 0.0;
  double traceInv = 0.0;

  for( int j=0; j < N; j++)
    { 
      double gj = W[j]*W[j];
      trace += gj;
      traceInv += 1.0/gj;
    }
  
  for ( int j=0; j < N; j++)
    {
      double gj = W[j]*W[j];
      double lj = gj/trace;
      double lj_1 = 1.0/gj/traceInv;
      S1 += lj*log(lj);
      S2 += lj_1*log(lj_1);
    }
  double c = S1 + S2;
  //printf("S1:  %f\n",S1);
  //printf("S2:  %f\n",S2);
  
  std::ofstream KLfile;
  KLfile.open(argv[3]);
  KLfile << c << std::endl;
  KLfile.close();
  */ 

  /* in principle, these must be divided by Ns to make true eigenvalues of the covariance matrix */
  for ( int i=0; i < Ns; i++)
    {
      if ( i == Ns -1)
	mypcafile << W[i]*W[i] << std::endl;
      else
	mypcafile << W[i]*W[i] << " ";
    }

  for ( int i=0; i< Ns; i++)
    {
    for ( int j=0; j< Nm; j++)
      for ( int k =0; k<Dimension; k++)
	mypcafile << U[j*Dimension+k][i] << " ";
    mypcafile << std::endl;
    }
  
  mypcafile.close();
  
  return EXIT_SUCCESS;
}
  
