/*
Author: Ali Gooya <ali dot gooya at gmail dot com>
*/
#include "TypeDefinitions.h"
#include "math.h"
#include <iostream>
#include <fstream>
#include <time.h>
#include <unistd.h>

double  ProjectShape(vnl_vector<double> *shape, int nb_modes, vnl_vector<double> *xBar, vnl_vector<double> *eigenValues, 
		     vnl_matrix<double> *modesMatrix, vnl_vector<double> *alignedShape, vnl_vector<double> *b)
{
  
  *b = (*modesMatrix)*( *shape - *xBar);

  for ( int i = 0; i < nb_modes; i++)
    {
      if ( (*b)[i] <= -3.0 * sqrt((*eigenValues)[i]))
	(*b)[i] = -3.0 * sqrt((*eigenValues)[i]);
      else if ( 3.0 * sqrt((*eigenValues)[i]) <= (*b)[i] )
	(*b)[i] = 3.0 * sqrt((*eigenValues)[i]);
    }
  
  *alignedShape = (*modesMatrix).transpose()*(*b) + *xBar;
  vnl_vector<double> dist = *alignedShape - *shape;
  double d = dist.two_norm()/((double)dist.size());
  return d;
}
void print_help()
{
      std::cerr << std::endl;
      std::cerr << "Demonstrates variations for a given input PCA." << std::endl;
      std::cerr << "The PCA file must be generated by ComputePCA." << std::endl;
      std::cerr << "Usage: demoVariations " << std::endl;
      std::cerr << " -p pcaModelFileName " << std::endl;
      std::cerr << " -b bVectorsFileName " << std::endl;
      std::cerr << "Options: " << std::endl;
      std::cerr << " -o outputGenericFileName (default: var)" << std::endl;
      std::cerr << std::endl;
}
 
int main( int argc, char *argv[] )
{
  int c = 0;
  int rOpts = 0;
  char bVectorsFileName[1000];
  char pcaModelFileName[1000];
  char outputGenericFileName[1000]="var";
  while (( c=getopt (argc, argv, "p:o:b:")) != -1)
    switch (c)
      {
      case 'p':
	sprintf(pcaModelFileName,"%s",optarg);
	++rOpts;
	break;
      case 'b':
	sprintf(bVectorsFileName,"%s",optarg);
	++rOpts;
	break;
      case 'o':
	sprintf(outputGenericFileName,"%s",optarg);
	break;
      case '?':
	fprintf (stderr, "Unknown option %s.\n", optarg);
	return 1;
      default:
	abort ();
      }
  if (rOpts != 2)
    {
      print_help();
      return 1;
    }

  std::ifstream bVectorsFile;
  bVectorsFile.open(bVectorsFileName);
  std::vector< vnl_vector<double> > bVectors;
  std::string line;
  while(std::getline(bVectorsFile, line))
    {
      std::stringstream linestream(line);
      double val;
      std::vector< double> lv;
      while(linestream >> val)
	{
	  lv.push_back(val);
	}
       vnl_vector<double> b;
      b.set_size(lv.end() -lv.begin());
      for(int i=0; i< lv.end() -lv.begin(); i++)
	b[i] = lv[i];
      bVectors.push_back(b);

    }
  int nb_vectors = bVectors.end() - bVectors.begin();
  
  std::vector< std::ofstream *> outFilesVector(nb_vectors);

  
  for (int i=0; i< nb_vectors; i++)
    {
      char name[1000];
      outFilesVector[i] = new std::ofstream[1];
      sprintf(name,"%s_%03d.txt",outputGenericFileName,i+1);
      outFilesVector[i]->open(name);
    }
  

  int nb_modes = bVectors[0].size();
  //std::cout << "nb_modes: " << nb_modes << std::endl;
  //now read PCA model
  std::ifstream pcaModel;
  pcaModel.open(pcaModelFileName);
  int n;
  pcaModel >> n;
  //std::cout << "n: " << n << std::endl;
  vnl_vector<double> xBar((n-1)*Dimension);
  pcaModel >> xBar;
  pcaModel.ignore(1000000, '\n');
  //std::cout << "xBar: " << xBar << std::endl;
  vnl_vector<double> eigenValues(nb_modes);
  pcaModel >> eigenValues ;
  pcaModel.ignore(1000000, '\n');
  //std::cout << eigenValues << std::endl;
  vnl_matrix<double> modesMatrix(nb_modes, (n-1)*Dimension);
  pcaModel >> modesMatrix;
  pcaModel.close();
  int m = 11;
  //std::cout << modesMatrix.get_row(0) << std::endl;
  std::vector< vnl_vector<double> > shapesVector( nb_vectors );

  for ( unsigned int k = 0; k < nb_vectors; k++ )
    {
      vnl_vector< double > b = bVectors[k];
      for (int i = 0;i< nb_modes; i++)
      	b[i] *= sqrt(eigenValues[i]/(double)m); ///(float)n);
      //std::cout << b << std::endl;
      shapesVector[k] = modesMatrix.transpose()*b + xBar;
      //shapesVector[k] = xBar;
      for (int i =0; i< shapesVector[k].size(); i += Dimension)
	{
	  VNLPointType vnlp;
	  for (int j =0; j< Dimension; j++)
	    vnlp[j] = shapesVector[k][i+j];
	  *(outFilesVector[k]) <<  vnlp << std::endl;

	}	    
      
      
      outFilesVector[k]->close();
    }

  
  return EXIT_SUCCESS;
}


