#include <fstream>
#include "TypeDefinitions.h"
#include <unistd.h>

void print_help()
{
  std::cerr << std::endl;
  std::cerr << "Usage: distanceOfPointSets " << std::endl; 
  std::cerr << " -i input point set list " << std::endl;
  std::cerr << " -r reference point set list " << std::endl;
  std::cerr << " -n number of point set pairs to be considered " << std::endl;
  std::cerr << "Options: " << std::endl;
  std::cerr << " -o computed distances (r2i) file name (default:distsR2I.txt) " << std::endl;
  std::cerr << " -O computed distances (i2r) file name (default:distsI2R.txt) " << std::endl;
  std::cerr << " -m computed max distances file name (default:maxDists.txt) " << std::endl;
  std::cerr << std::endl;
}

int main(int argc, char *argv[])
{

  char inputList[1000];
  char refList[1000];
  char r2iDistsFileName[1000]="distsR2I.txt";
  char maxDistsFileName[1000]="maxDists.txt";
  char i2rDistsFileName[1000]="distsI2R.txt";
  int nbFiles = 0;
  int rOpts = 0;
  int c = 0;
  while ((c = getopt (argc, argv, "i:r:o:m:n:O:")) != -1)
    switch (c)
      {
      case 'i':
	sprintf(inputList,"%s",optarg);
	++rOpts;
	break;
      case 'n':
	nbFiles = atoi(optarg);
	++rOpts;
	break;
      case 'r':
	sprintf(refList,"%s",optarg);
	++rOpts;
	break;
      case 'o':
	sprintf(r2iDistsFileName,"%s",optarg);
	break;
      case 'm':
	sprintf(maxDistsFileName,"%s",optarg);
	break;
      case 'O':
	sprintf(i2rDistsFileName,"%s",optarg);
	break;
      case '?':
	fprintf (stderr, "Unknown option %s.\n", optarg);
	return 1;
      default:
	abort ();
      }
 if (rOpts != 3 )
    {
      print_help();
      return 1;
    }
	
 //figuring out the input point set files
 std::fstream inputPointSetList;
 inputPointSetList.open(inputList);
 std::vector<char *> inputPointSetsNamesVector( nbFiles);
 for ( unsigned int k = 0; k < nbFiles; k++ )
    {
      inputPointSetsNamesVector[k] = new char[1000];
    }
  //Next, we make input pointset
  //Something to keep input the point sets
 typedef vnl_vector_fixed<float, Dimension> inputPointType;
 typedef std::vector<inputPointType> pointsVectorType;
 std::vector<pointsVectorType>    inputPointSetsVector( nbFiles );


 for ( unsigned int k = 0; k < nbFiles; k++ )
    {
      inputPointSetList >> inputPointSetsNamesVector[k];
      std::fstream myfile;
      myfile.open(inputPointSetsNamesVector[k]);
      while( !myfile.eof() )
	{
	  std::string str;
	  std::getline(myfile,str);
	  if (!str.empty()) //a point
	    {
	      std::istringstream tmp(str);
	      inputPointType point;
	      tmp >> point;
	      inputPointSetsVector[k].push_back(point);
	    }
	}
      myfile.close();
    }
 //figuring out the reference point set files
 std::fstream referencePointSetList;
 referencePointSetList.open(refList);
 std::vector<char *> referencePointSetsNamesVector( nbFiles);
 for ( unsigned int k = 0; k < nbFiles; k++ )
    {
      referencePointSetsNamesVector[k] = new char[1000];
    }


 std::vector<pointsVectorType>    referencePointSetsVector( nbFiles );
 for ( unsigned int k = 0; k < nbFiles; k++ )
    {
      referencePointSetList >> referencePointSetsNamesVector[k];
      std::fstream myfile;
      myfile.open(referencePointSetsNamesVector[k]);
      while( !myfile.eof() )
	{
	  std::string str;
	  std::getline(myfile,str);
	  if (!str.empty()) //a point
	    {
	      std::istringstream tmp(str);
	      inputPointType point;
	      tmp >> point;
	      referencePointSetsVector[k].push_back(point);
	    }
	}
      myfile.close();
    }

 //now computing distances between point set pairs 
 vnl_vector<double> r2iDists(nbFiles);
 vnl_vector<double> i2rDists(nbFiles);
 vnl_vector<double> maxDists(nbFiles);

 for ( unsigned int k = 0; k < nbFiles; k++ )
    {
      int inpN = inputPointSetsVector[k].end() - inputPointSetsVector[k].begin(); //number of points for input point set
      int refN = referencePointSetsVector[k].end() - referencePointSetsVector[k].begin(); //number of points for reference point set
      double * i2rD = new double[inpN];
      double * r2iD = new double[refN];
      for (int i=0; i< inpN; i++)
	{
	  double minDist = 1e32;
	  inputPointType iP = inputPointSetsVector[k][i];
	  inputPointType matchDiff;
	  for (int r=0; r< refN; r++)
	    {
	      inputPointType rP = referencePointSetsVector[k][r];
	      inputPointType dP = rP - iP;
	      double d = dP.magnitude();
	      //double d=0.0;
	      //for (int j=0; j< Dimension-1; j++) d += dP[j]*dP[j];
	      if ( d < minDist)
		{
		  minDist=d;
		  matchDiff = dP;
		}
	    }
	  i2rD[i] = minDist;
	  //i2rD[i] = matchDiff[Dimension-1]*matchDiff[Dimension-1];
	}
 
     //now finding maximal and sum of minimals
      double S = 0.0;
      double maxD = 0.0;
      for (int i=0; i< inpN; i++)
	{
	  S += i2rD[i];
	  if( maxD < i2rD[i])
	    maxD = i2rD[i];
	}
      //i2rDists[k] = vcl_sqrt(S/inpN/inpN);
      i2rDists[k] = S/inpN;
      maxDists[k] = maxD;

     //now changing the order
      for (int r=0; r< refN; r++)
	{
	  double minDist = 1e32;
	  inputPointType rP = referencePointSetsVector[k][r];
	  inputPointType matchDiff;
	  for (int i=0; i< inpN; i++)
	    {
	      inputPointType iP = inputPointSetsVector[k][i];
	      inputPointType dP = rP - iP;
	      double d = dP.magnitude();
	      //double d=0.0;
	      //for (int j=0; j< Dimension-1; j++) d += dP[j]*dP[j];
	      if ( d < minDist)
		{
		  minDist=d; 
		  matchDiff = dP;
		}
	    }
	  r2iD[r] = minDist;
	  //r2iD[r] = matchDiff[Dimension-1]*matchDiff[Dimension-1];
	}
      //now finding maximal and sum of minimals
      S = 0.0;
      maxD = 0.0;
      for (int r=0; r< refN; r++)
	{
	  S += r2iD[r];
	  if ( maxD < r2iD[r])
	    maxD = r2iD[r];
	}

      //r2iDists[k] = vcl_sqrt(S/refN/refN); 
      r2iDists[k] = S/refN ; 
      if ( maxDists[k] <  maxD ) maxDists[k] = maxD;

      delete i2rD;
      delete r2iD;
    }
 
 std::cout << "r2i distances: " << r2iDists << std::endl;
 //std::cout << "i2r distances: " << i2rDists << std::endl;
 std::cout << "max distances: " << maxDists << std::endl;

 // now  writing results to the files
 std::ofstream r2iDistsFile;
 r2iDistsFile.open(r2iDistsFileName);
 std::ofstream i2rDistsFile;
 i2rDistsFile.open(i2rDistsFileName);
 std::ofstream maxDistsFile;
 maxDistsFile.open(maxDistsFileName);

 for ( int k =0; k < nbFiles; k++)
   {
     r2iDistsFile << r2iDists[k] + i2rDists[k]<< std::endl;
     i2rDistsFile << i2rDists[k] + r2iDists[k]<< std::endl;
     maxDistsFile << maxDists[k] << std::endl;
   }
 r2iDistsFile.close();
 i2rDistsFile.close();
 maxDistsFile.close();

 for ( unsigned int k = 0; k < nbFiles; k++ )
    {
      delete referencePointSetsNamesVector[k];
      delete inputPointSetsNamesVector[k];
    }

 return 0;
}
