//#include "libqp.h"
//#include "libqp_splx.c"
#include "libqp_gsmo.c"
static uint32_t nVars;
static double *mat_H = NULL;

const double *get_col_of_mat_H( uint32_t i )
{
  return( &mat_H[ nVars*i ] );
}

void UpdateConstrainedPriorsVector(PriorsVectorType * uncPriorsVector, PriorsVectorType * priorsVector)
{
  double *diag_H = NULL;
  double *vec_f = NULL;
  //double *vec_b = NULL;
  //uint32_t *vec_I = NULL;
  //uint8_t *vec_S = NULL;
  double *vec_x = NULL;
  libqp_state_T exitflag;
  uint32_t MaxIter = 0xFFFFFFFF;
  //double TolAbs = 0.0;                  
  //double TolRel = 1e-9;
  //double QP_TH = -LIBQP_PLUS_INF;
  double *a = NULL;
  double *LB = NULL;
  double *UB = NULL;
  double TolKKT = 0.0;

  PriorsVectorType::iterator priorsVectorIt;
  priorsVectorIt = priorsVector->begin();

  PriorsVectorType::iterator uncPriorsVectorIt;
  uncPriorsVectorIt = uncPriorsVector->begin();
  nVars = uncPriorsVector->end() - uncPriorsVector->begin();
  //this is to check initialy if we really need optimization 
  bool flag = false;
  double uncSum = 0.0;

  while(uncPriorsVectorIt != uncPriorsVector->end())
    {
      double f = (*uncPriorsVectorIt);
      uncSum += f;
      if ( f < 0.0 || 1.0 < f )
	flag = true;
      ++uncPriorsVectorIt;
    }
  //std::cout << "uncSum: " << uncSum << std::endl;
  //std::cout << "flag_min_max:" << flag << std::endl;
  /*
  if ( (uncSum - 1.0) )
    {
      std::cout << "uncSum: " << uncSum << std::endl;
      flag = true;
    }
  
  std::cout << "flag_sum:" << flag <<  std::endl;
  */
  if (flag == false) // no need for optimization
    {
      //std::cout << "priors meet KKT, no need for optimization." << std::endl;
      uncPriorsVectorIt = uncPriorsVector->begin();
      priorsVectorIt = priorsVector->begin();
      while(uncPriorsVectorIt != uncPriorsVector->end())
	{
	  *priorsVectorIt = (*uncPriorsVectorIt);
	  ++uncPriorsVectorIt;
	  ++priorsVectorIt;
	}
	  
      return;
    }
  
  //This is important, initialize to feasible solution
  vec_x = (double *)calloc(nVars,sizeof(double));
  for(int i=0; i < nVars; i++)
    vec_x[i] = 1.0/((double)nVars);
   
  vec_f = (double *)calloc(nVars,sizeof(double));
  uncPriorsVectorIt = uncPriorsVector->begin();
  while(uncPriorsVectorIt != uncPriorsVector->end())
    {
      int j = uncPriorsVectorIt - uncPriorsVector->begin();
      vec_f[j] = 0.0 - (*uncPriorsVectorIt);
      ++uncPriorsVectorIt;
    }

  mat_H = (double*)calloc(nVars*nVars,sizeof(double));
  for(int i=0; i < nVars; i++)
    for(int j=0; j < nVars; j++)
      {
	if( i == j ) mat_H[i*nVars+j] = 1.0;
	else mat_H[i*nVars+j] = 0.0;
      }

  diag_H = (double*)calloc(nVars,sizeof(double));
  for(int i=0; i < nVars; i++)
    diag_H[i] = 1.0;

   
  double b = 1.0;
  a = (double *)calloc(nVars,sizeof(double));
  UB = (double *)calloc(nVars,sizeof(double));
  LB = (double *)calloc(nVars,sizeof(double));
  for(int i=0; i < nVars; i++)
    {
      a[i] = 1.0;
      UB[i] = 1.0;
      LB[i] = 0.0;
    }
  
  exitflag = libqp_gsmo_solver(get_col_of_mat_H, diag_H, vec_f, a, b, LB, UB, vec_x, nVars, MaxIter, TolKKT, NULL);
  /*
  vec_b = (double *)calloc(1,sizeof(double));
  *vec_b = 1.0;

  vec_I = (uint32_t *)calloc(nVars,sizeof(uint32_t));
  for(int i=0; i < nVars; i++)
    vec_I[i] = i;

  vec_S = (uint8_t *)calloc(nVars,sizeof(uint8_t));
  for(int i=0; i < nVars; i++)
    vec_S[i] = 0;
  

  exitflag = libqp_splx_solver(get_col_of_mat_H, diag_H, vec_f, vec_b,
                               vec_I, vec_S, vec_x, nVars, MaxIter,
                               TolAbs, TolRel, QP_TH, NULL);

  */
  /*
  printf("exitflag: %d\n", exitflag.exitflag);
  printf("6\n");
  printf("QP 1 1\n");
  printf("%.12f\n",exitflag.QP);

  printf("QD 1 1\n");
  printf("%.12f\n",exitflag.QD);

  printf("nIter 1 1\n");
  printf("%d\n",exitflag.nIter);

  printf("exitflag 1 1\n");
  printf("%d\n",exitflag.exitflag);

  */

  double sum = 0.0;
  double priorSum = 0.0;
  double min = 1.0;
  double max = 0.0;
  while(priorsVectorIt != priorsVector->end())
    {
      int j = priorsVectorIt - priorsVector->begin();
      
      (*priorsVectorIt) = vec_x[j];
      priorSum += (*priorsVectorIt);
      if ((*priorsVectorIt) < 0.0 )
	std::cout << "Negative prior value is found: " << (*priorsVectorIt) << std::endl;
      else if ((*priorsVectorIt) > 1.0 )
	std::cout << "Prior value is bigger than 1.0: " << (*priorsVectorIt) << std::endl;
      else if ( (*priorsVectorIt) == 0.0 )
	sum += 1;
      else if ( (*priorsVectorIt) < min )
	min = (*priorsVectorIt);
      else if ( (*priorsVectorIt) > max )
	max = (*priorsVectorIt);
     ++priorsVectorIt;
    }
  //std::cout << "sparsity: " << sum/(float)nVars*100.0 << std::endl;
  //std::cout << "remaining alive points: " << nVars - sum << std::endl;
  /*std::cout << "min prior: " << min << std::endl;
  std::cout << "max prior: " << max << std::endl;
  std::cout << "prior sum: " << priorSum << std::endl;*/
  
  

  free(diag_H);
  free(mat_H);
  //free(vec_I);
  free(vec_f);
  //free(vec_b);
  //free(vec_S); 
  free(vec_x); 
  free(a);
  free(UB);
  free(LB);



}
