function [d dMax] = crossValidateDistances(inputMeshFilesList, outputDirName, estPrior, estAlpha, N,S,Nm,z,nIter,R,P_S,dx,scale)
% N is the total number of shapes
% S is the number of leave out samples
% z is the desired sparsity
% Nm is the number of model points (used only for classic model)
% estPrior set 1 if classic model is desired
% estAlpha set 1 if estimation of alpha is desired
% nIter is the number of EM iterations for classic model
% R is the variance reduction factor in the classic model
% P_S if set surface distances will be measured using provided templateImage.
% scale is the kernel for RBF is P_S is set

%first create a local dir to save results
mkdir(outputDirName);
cd(outputDirName);

%then read available mesh file names and save them in an array
fidMeshFiles = fopen(inputMeshFilesList,'r');
for i = 1:N
    inputMeshNames(i,:) = fscanf(fidMeshFiles, '%s',1);
end
fclose(fidMeshFiles);
%

numIter = N/min(S,N-S);
%dists = zeros(S, numIter);
%currFiles = sprintf('%s%s',pwd,'/testMesh_%02d.txt');
for  i = 1:numIter
    if ( S < N-S)
         %create a name for a file containing training set names
         trainingSamples = sprintf('%s/trainingSamples_%02d.txt',pwd,i);
         fidTrain = fopen(trainingSamples,'w');
         %populating it
         %first group
         sTrain1 = 1;
         eTrain1 = (numIter - i)*S;
         while( ge(eTrain1,sTrain1) )
            trainMesh = inputMeshNames(sTrain1,:);
            fprintf(fidTrain, '%s\n', trainMesh);
            sTrain1 = sTrain1 + 1;
         end
         %second group
         sTrain2 = (numIter - i  + 1)*S + 1;
         eTrain2 = N;
         while( ge(eTrain2,sTrain2) )
            trainMesh = inputMeshNames(sTrain2,:);
            fprintf(fidTrain, '%s\n', trainMesh);
            sTrain2 = sTrain2 + 1;
         end
         fclose(fidTrain);
         %create a name for a file containing test set names
         testSamples = sprintf('%s/testSamples_%02d.txt',pwd,i);
         fidTest = fopen(testSamples,'w');
         sTest = (numIter - i)*S + 1;
         eTest = (numIter - i + 1)*S;
         while( ge(eTest,sTest) )
            testMesh = inputMeshNames(sTest,:);
            fprintf(fidTest,'%s\n',testMesh);
            sTest = sTest + 1;
         end
         fclose(fidTest);
    else
          %create a name for a file containing training set names
         trainingSamples = sprintf('%s/trainingSamples_%02d.txt',pwd,i);
         fidTrain = fopen(trainingSamples,'w');
         %populating it
         sTrain = (numIter - i)*(N-S) + 1;
         eTrain = (numIter - i + 1)*(N-S);
         while( ge(eTrain,sTrain) )
            trainMesh = inputMeshNames(sTrain,:);
            fprintf(fidTrain,'%s\n',trainMesh);
            sTrain = sTrain + 1;
         end
         fclose(fidTrain);
         %create a name for a file containing test set names
         testSamples = sprintf('%s/testSamples_%02d.txt',pwd,i);
         fidTest = fopen(testSamples,'w');

         %first group
         sTest1 = 1;
         eTest1 = (numIter - i)*(N-S);
         while( ge(eTest1,sTest1) )
            testMesh = inputMeshNames(sTest1,:);
            fprintf(fidTest, '%s\n', testMesh);
            sTest1 = sTest1 + 1;
         end
         %second group
         sTest2 = (numIter - i  + 1)*(N-S) + 1;
         eTest2 = N;
         while( ge(eTest2,sTest2) )
            testMesh = inputMeshNames(sTest2,:);
            fprintf(fidTest, '%s\n', testMesh);
            sTest2 = sTest2 + 1;
         end
         fclose(fidTest);
    end
    
    model = sprintf('model_%02d.txt',i);
    trainVir = sprintf('trainVir_%02d',i);
    trainAlg = sprintf('trainAlg_%02d',i);
    trainVirList = sprintf('trainVirList_%02d.txt',i);
    trainVirCorr = sprintf('trainVirCorr_%02d.txt',i);
    trainAlgList = sprintf('trainAlgList_%02d.txt',i);
    trainLogFile = sprintf('trainLogFile_%02d.txt',i);
    trainTran = sprintf('trainTran_%02d',i);
    trainEntFile = sprintf('trainEntFile_%02d.txt',i); 
    
    if ( estPrior)     
            if ( estAlpha == 0)
                com = sprintf('sparseModel -n %d -i %s -M %s -z %f -g %s -T %s -a %s -V %s -C %s -A %s -L %s -c %d -p %d -N 4000 -r 50 -x %s', N - S, trainingSamples, model, z, trainVir,...
                    trainTran, trainAlg, trainVirList, trainVirCorr, trainAlgList, trainLogFile, 100, 0, trainEntFile);
            else
                com = sprintf('sparseModel -n %d -i %s -M %s -z %f -g %s -T %s -a %s -V %s -C %s -A %s -L %s -c %d -p %d -N 4000 -r 50 -x %s', N - S, trainingSamples, model, z, trainVir,...
                    trainTran, trainAlg, trainVirList, trainVirCorr, trainAlgList, trainLogFile, 100, 1, trainEntFile);
            end
    %this line is to check classic model
    else
        com= sprintf('sparseModel -n %d -i %s -M %s -z %f -g %s -T %s -a %s -V %s -C %s -A %s -L %s -c %d -s %d -N %d -r %d -R %f', N - S, trainingSamples, model, z, trainVir,...
            trainTran, trainAlg, trainVirList, trainVirCorr, trainAlgList, trainLogFile, 100, 0, Nm, nIter, R);
    end
    %system(com,'-echo');
    
    
    testVir = sprintf('testVir_%02d',i);
    testAlg = sprintf('testAlg_%02d',i);
    testVirList = sprintf('testVirList_%02d.txt',i);
    testVirCorr = sprintf('testVirCorr_%02d.txt',i);
    testAlgList = sprintf('testAlgList_%02d.txt',i);
    testLogFile = sprintf('testLogFile_%02d.txt',i);
    testTran = sprintf('testTran_%02d',i);
    %testProbFile = sprintf('testProbFile_%02d.txt',i);
 
    
    %now bringing test set into alignment
    if ( estPrior) 
        com = sprintf('sparseModel -n %d -i %s -M %s -m %d -g %s -T %s -a %s -V %s -C %s -A %s -L %s -r 50', S, testSamples, model,0,testVir,...
            testTran, testAlg, testVirList, testVirCorr, testAlgList, testLogFile);
    %%% this line is to check classic model
    else
        com = sprintf('sparseModel -n %d -i %s -M %s -m %d -g %s -T %s -a %s -V %s -C %s -A %s -L %s -s %d -r %d', S, testSamples, model,0,testVir,...
            testTran, testAlg, testVirList, testVirCorr, testAlgList, testLogFile, 0, nIter);
    end
    %system(com,'-echo');
    
    
    %now computing pca
    pcaModel = sprintf('pcaModel_%02d.txt',i);
    com =sprintf('computePCA -i %s -o %s ', trainVirCorr,pcaModel);        
    %system(com,'-echo');
    
    %now projecting test set into pca
    testPrjList=sprintf('testPrjList_%02d.txt',i);
    testBVectors=sprintf('testBVectors_%02d.txt',i);
    com = sprintf('projectVirtuallyAlignedShapesToModel -n %d -i %s -p %s -m %d -o %s -b %s',S,testVirList,pcaModel, N-S, testPrjList,testBVectors);
    %system(com, '-echo');
    
    if ( isempty(P_S) )
        %now computing point set distances
        dists=sprintf('dists_%02d.txt',i);
        maxDists=sprintf('maxDists_%02d.txt',i);
        com = sprintf('distanceOfPointSets -r %s -i %s -n %d -o %s -m %s', testPrjList, testAlgList, S, dists, maxDists);
        %system(com,'-echo'); 
        %now saving computed probabilites and distances
        fidDists = fopen(dists,'r');
        for j = 1:S
            d(j,i) = fscanf(fidDists,'%f',1);
        end
        fclose(fidDists);
        fidMaxDists = fopen(maxDists,'r');
        for j = 1:S
            dMax(j,i) = fscanf(fidMaxDists,'%f',1);
        end
        fclose(fidMaxDists);
    else
         %now computing surface distances
        surfDistsList=sprintf('surfDistsList_%02d.txt',i);
        com = sprintf('ComputeSurfaceDistanceForMeshListS.py -r %s -i %s -d %s -n %d -x %f -y %f -z %f -k %d -s %f', testAlgList,testPrjList,surfDistsList,S,dx,dx,dx,0,scale);
        system(com,'-echo'); 
        %now saving computed probabilites and distances
        fidSurf = fopen(surfDistsList,'r');
        for j = 1:S
            d(j,i) = fscanf(fidSurf,'%f',1);
        end
        fclose(fidSurf);
    end
    
    
     
      
end
cd ../


end
