#!/usr/bin/python
import getopt, sys, os, commands, shutil

def usage():
    print "\nUsage:\n"
    print "Converts point sets in a list to level images and computes distances."
    print "\tComputeSurfaceDistanceForMeshListS -i input_pointset_list -r reference_pointset_list -d output_distance_list -n number_of_meshes \n"
    print "\tOptions:"
    print "\t-s scale used for imge generation"
    print "\-x -y -z sampling distances for image generation"
    print "\t-k set if generated images should be kept (defualt:1)"
    print "\t-t template image"
    #print "\t-l lower threshold level object extraction (default: 0)"
    #print "\t-u upper threshold level object extraction (default: 255)"
    
    
    
def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hvi:r:d:n:t:s:l:u:k:x:y:z:", [])
    except getopt.GetoptError, err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        usage()
        sys.exit(2)
    outputDir = os.getcwd()
    verbose = False
    outputDir = os.path.realpath(outputDir)
    
    outputImageList = ""
    testFileList = ""
    refFileList = ""
    templateImage = ""
    refImage = ""
    rOpts = 0
    n = 0
    s = " 4 "
    lowT = " 0 "
    upT = " 255 "
    k = 1
    x = " 2 "
    y = " 2 "
    z = " 2 "
    for o, a in opts:
        if o == "-v":
            verbose = True
        elif o in ("-h", "--help"):
            usage()
            sys.exit()
        #elif o in ("-o"):
        #    outputImageList = a
        #    rOpts +=1
        elif o in ("-i"):
            testFileList = a
            rOpts +=1
        elif o in ("-r"):
            refFileList = a
            rOpts +=1
        elif o in ("-n"):
            n = int(a)
            rOpts +=1
        elif o in ("-d"):
            outputDiceList = a
            rOpts +=1
        elif o in ("-t"):
            templateImage = a
        elif o in ("-s"):
            s = a
        elif o in ("-x"):
            x = a
        elif o in ("-y"):
            y = a
        elif o in ("-z"):
            z = a
        elif o in ("-l"):
            lowT = a
        elif o in ("-u"):
            upT = a
        elif o in ("-k"):
            k = int(a)
        else:
            assert False, "unhandled option"
   
    if (rOpts != 4):
        usage()
        print "Please specify all required inputs.\n"
        sys.exit(2)

    testFileNames = " "
    if os.path.exists(testFileList):
        testList = open(testFileList)
        testFileNames = testList.readlines()
    else:
        print "Can not find input file list: "+testFileList
    refFileNames = " "
    if os.path.exists(refFileList):
        refList = open(refFileList)
        refFileNames = refList.readlines()
    else:
        print "Can not find input file list: "+testFileList


 
    #if not os.path.exists(templateImage):
    #    print "Can not find templateImage: "+templateImage


    cnt=0
    ofile = open(outputDir+"/"+outputDiceList,'w')


    for eachF in testFileNames:
        if not (os.path.exists(eachF.strip())):
            print "Can not find input mesh: " + eachF
            continue
        eachR = refFileNames[cnt]
        if not (os.path.exists(eachR.strip())):
            print "Can not find input mesh: " + eachR
            continue
        #make a reference image
        refImage=outputDir+"/"+os.path.splitext(os.path.basename(eachR))[0]+".nii.gz"
        if not os.path.exists(refImage):
            com="pointSetToImage -i "+eachR.strip()+" -o "+refImage+" -x "+x+" -y "+y+" -z "+z+" -s "+s+" -g 0"
            print com
            os.system(com)
        #making a binary image 
        thrRefImage=outputDir+"/"+os.path.splitext(os.path.basename(eachR))[0]+"_thr.nii.gz"
        com="3dcalc -a "+refImage+" -prefix "+thrRefImage+" -expr 'step(a)' "
        os.system(com)
        print "finding the edges in the binarized reference image ..."
        edgeRefImage=outputDir+"/"+os.path.splitext(os.path.basename(eachR))[0]+"_edge.nii.gz"
        com = "3dcalc -a "+thrRefImage+" -prefix "+edgeRefImage+ " -b  a+i -c a-i -d a+j -e a-j -f a+k -g a-k -expr 'a*amongst(0,b,c,d,e,f,g)'"  
        os.system(com)
        #eachR = os.path.realpath(eachR)

        eachF = os.path.realpath(eachF)
        eachO = outputDir+"/"+os.path.splitext(os.path.basename(eachF))[0]+".nii.gz"
        #outputList.write(eachO+"\n")
        if not os.path.exists(eachO):
            #com="PointSetToImage "+eachF.strip()+" "+eachO+" "+templateImage+" "+s
            #com="pointSetToImage -i "+eachF.strip()+" -o "+eachO+" -t "+templateImage+" -s "+s+" -g 0"
            com="pointSetToImage -i "+eachF.strip()+" -o "+eachO+" -t "+refImage+" -s "+s+" -g 0"
            print com
            os.system(com)
        prod1 = outputDir+"/"+os.path.splitext(os.path.basename(eachF))[0]+"_prodRefEdge1.nii.gz"
        com = "3dcalc -a "+edgeRefImage+" -b "+eachO+" -prefix "+prod1+" -expr 'step(a)*(b*step(b)-b*step(-b))' "
        print com
        os.system(com)
        print "computing first distance "
        com = "fslstats "+prod1+" -M "
        stat, val= commands.getstatusoutput(com)
        dist1 = float(val)
        thrEachO = outputDir+"/"+os.path.splitext(os.path.basename(eachF))[0]+"_thr.nii.gz"
        #com = "fslmaths "+eachO+" -thr "+lowT+" -uthr "+upT+" "+thrEachO
        com="3dcalc -a "+eachO+" -prefix "+thrEachO+" -expr  'step(a)' "
        os.system(com)
        print "finding the edges in the binarized input mesh ..."
        edgeEachO = outputDir+"/"+os.path.splitext(os.path.basename(eachF))[0]+"_edge.nii.gz"
        com = "3dcalc -a "+thrEachO+" -prefix "+edgeEachO+ " -b  a+i -c a-i -d a+j -e a-j -f a+k -g a-k -expr 'a*amongst(0,b,c,d,e,f,g)'"  
        os.system(com)
        prod2 = outputDir+"/"+os.path.splitext(os.path.basename(eachF))[0]+"_prodRefEdge2.nii.gz"
        com = "3dcalc -a "+edgeEachO+" -b "+refImage+" -prefix "+prod2+" -expr 'step(a)*(b*step(b)-b*step(-b))' "
        os.system(com)
        print "computing second distance "
        com = "fslstats "+prod2+" -M "
        stat, val= commands.getstatusoutput(com)
        dist2 = float(val)
        print "total distance: "+str(dist1+dist2)
        ofile.writelines(str(dist1+dist2)+"\n")
        os.remove(prod1)
        os.remove(prod2)
        os.remove(edgeEachO)
        os.remove(thrEachO)
        os.remove(thrRefImage)
        os.remove(edgeRefImage)
        if not k:
            os.remove(eachO)
            os.remove(refImage)
             
        cnt+=1
        if (n <= cnt):
            break
        
    #outputList.close()
    ofile.close()
    #os.remove(thrRefImage)
    #os.remove(edgeRefImage)
     
 

if __name__ == "__main__":
    main()
