/*
Author: Ali Gooya <ali dot gooya at gmail dot com>
*/
#include "TypeDefinitions.h"
#include <flann/flann.hpp> 
#include "math.h"
#include "libQpMin.cxx"
#include "kmeans.h"
#include "vnl/vnl_inverse.h"
#include <vnl/algo/vnl_matrix_inverse.h>
#include <vnl/vnl_matrix.h>
#include <vnl/vnl_trace.h>
#include <iostream>
#include <fstream>
#include <time.h>
#include "cluster.c"
#include <vnl/vnl_diag_matrix.h>
#include <time.h>
#include "Utilities.cxx"
#include <unistd.h>
#include <libgen.h>


void InitializeGmmMeans(PointSetVectorType * pointSetVector, PointSetType *initialMeanPointSet, int k, int npass)
{

  int n = 0;
  PointSetVectorType::iterator shapeIt = pointSetVector->begin();
  
  while ( shapeIt !=  pointSetVector->end() )
    {
      n += ((*shapeIt)->GetNumberOfPoints());
      ++shapeIt;
    }
  int m = GMMPointDimension;
 
  //pre-processing data
  vnl_vector_fixed<double, GMMPointDimension> ex;
  vnl_vector_fixed<double, GMMPointDimension> ex2;
  ex.fill(0);
  ex2.fill(0);

  shapeIt = pointSetVector->begin();
  int N = 0;
  while ( shapeIt !=  pointSetVector->end() )
    {
      int Ns = (*shapeIt)->GetNumberOfPoints();
      for (int i = 0; i < Ns; i++)
	{
	  VNLPointType shapeVNLPoint;
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  //std::cout << "shapeVNLPoint: " << (shapeVNLPoint) << std::endl;
	  for (int j = 0; j < m ; j++)
	    {
	      ex[j] += shapeVNLPoint[j];
	      ex2[j] += shapeVNLPoint[j]*shapeVNLPoint[j];
	    }
	  
	}
      N += Ns;
      ++shapeIt;
    }

  ex *= 1.0/(double)N;
  ex2 *= 1.0/(double)N;

  vnl_vector_fixed<double, GMMPointDimension> var = ex2;
  for(int j = 0; j < m; j++)
    {
      var[j] -= ex[j]*ex[j];
    }

  double ** data = (double**) malloc(n*sizeof(double*));
  int ** mask = (int**) malloc(n*sizeof(int*));
  int * clusterid = (int*) malloc(n*sizeof(int));
  for (int i =0 ; i < n; i++)
    {
      data[i] = (double*) malloc(m*sizeof(double));
      mask[i] = (int*) malloc(m*sizeof(int));
    }
  N = 0;
  shapeIt = pointSetVector->begin();
  while ( shapeIt !=  pointSetVector->end() )
    {
      int Ns = (*shapeIt)->GetNumberOfPoints();
      for (int i = 0; i < Ns; i++)
	{
	  VNLPointType shapeVNLPoint;
	  GetVNLPoint(*shapeIt, i , &shapeVNLPoint);
	  for (int j = 0; j < m ; j++)
	    {
	      data[ i + N][j] = (shapeVNLPoint[ j ] - ex[j])/vcl_sqrt(var[j]);
	      mask[i + N][j] = 1.0;
	    }
	}
      N = N + Ns;
      ++shapeIt;
    }
  char method = 'e';
  
  double * weight = (double*) malloc(m*sizeof(double));
  for (int i=0; i< m; i++) weight[i] = 1.0;
  //weight[m-1] = 0.0;
  double error = 0.0;
  int iffound = 0;
  char dist = 'e';
  //int npass = 5000;
  int transpose = 0;

  kcluster(k,n,m,data,mask,weight,transpose,npass,method,dist,clusterid,&error,&iffound);
  
  double ** cdata = (double**) malloc(k*sizeof(double*));
  int ** cmask = (int**) malloc(k*sizeof(int*));
  for (int i =0 ; i < k; i++)
    {
      cdata[i] = (double*) malloc(m*sizeof(double));
      cmask[i] = (int*) malloc(m*sizeof(int));
    }
  for ( int i =0; i <k; i++)
    for( int j =0; j<m; j++)
      cmask[i][j] = 1.0;
  
  method = 'm';
  getclustercentroids(k, n, m, data, mask, clusterid , cdata, cmask, 0, method);
  //std::cout << "initialization state: " << getclustercentroids(k, n, m, data, mask, clusterid , cdata, cmask, 0, method) << std::endl;

  for (int i=0;i<k;i++) {
    VNLPointType vnlp;
    for (int j=0;j<m;j++){
	//	printf("%f ",vcl_sqrt(var[j])*cdata[i][j]+ex[j]);
      vnlp[j] = vcl_sqrt(var[j])*cdata[i][j]+ex[j];
    }
	   // printf("\n");
    SetVNLPoint( initialMeanPointSet,i,&vnlp); 
  }
 
  std::cout << "initialization of model means done.." << std::endl;

  delete data;
  delete mask;
  //delete centerid;
  delete cdata;
  delete cmask;
  delete weight;
  delete clusterid;

}

void WriteModel(PointSetType * points,  PriorsVectorType * priorsVector, char * name, double * sigmaV)//, double *sigmaf)
{
  std::ofstream myfile;
  myfile.open(name);
  //std::cout << "writing points" << std::endl;
  PriorsVectorType::iterator priorsVectorIt = priorsVector->begin();
  for (int i=0; i<GMMPointDimension; i++)
    myfile << sigmaV[i] << " ";
  myfile <<std::endl;
  //myfile << (*sigma) << " " << (*sigmaf) << std::endl;
  

  PointsIteratorType it = points->GetPoints()->Begin();
  while (it != points->GetPoints()->End())
    {
      if ( it != points->GetPoints()->Begin())
	myfile << std::endl;
      
      VNLPointType vnlp;
      GetVNLPoint(points, it.Index() , &vnlp);
      myfile << vnlp ;
      myfile << " " << (*priorsVectorIt);
      ++priorsVectorIt;
      ++it;
    }
  myfile.close();

}



void print_help()
{

      std::cerr << std::endl;
      std::cerr << "Usage: sparseModel " << std::endl; 
      std::cerr << " -n number of input meshes " << std::endl;
      std::cerr << " -i input mesh list"  << std::endl;
      std::cerr << " -M GMM model file name. This can be an input or output file name (depending on option -m) " << std::endl;
      std::cerr << "Options: " << std::endl;
      std::cerr << " -N number of initial model points (default:1e4) " << std::endl;
      std::cerr << " -c number of initial k-means pass (default:1)" << std::endl;
      std::cerr << " -m set if estimation of model is desired (defualt:1)" << std::endl;
      std::cerr << " -v set if virtual correspondance is desired (default: 1)" << std::endl;
      std::cerr << " -s set if sparsity is desired (default:1)" << std::endl;
      std::cerr << " -z desired sparsity amount (default:0.5, must between 0 to 1 )" << std::endl;
      std::cerr << " -t set if estimation of transformation is desired (default:1)" << std::endl;
      std::cerr << " -r the maximum number of EM iterations (default: 100)" << std::endl;
      std::cerr << " -g virtual aligned mesh generic file name (default: vir)" << std::endl;
      std::cerr << " -T generic output transformations file name (default:tran )" << std::endl;
      std::cerr << " -a model aligned mesh file generic name (default: alg)" << std::endl;
      std::cerr << " -V virtual aligned mesh list file name (default: virList.txt)" << std::endl;
      std::cerr << " -C a file name to keep all virtually correspondant points(default: virCorr.txt)" << std::endl;
      std::cerr << " -A model aligned mesh list file name (default: algList.txt)" << std::endl;
      std::cerr << " -L logfile name (default: logfile.txt) " << std::endl;
      //std::cerr << " -p set if estimation of alpha is desired (default:0)" << std::endl;
      //std::cerr << " -x KL entropy file name (default:entorpy.txt) " << std::endl;
      //std::cerr << " -B bic value file name (default:bic.txt)" << std::endl;
      std::cerr << " -R variance reduction factor for classic model  (default:0.85)" << std::endl;
      
      std::cerr << std::endl;
}



int main( int argc, char *argv[] )
{

  int rOpts = 0;
  unsigned int nb_train=0;
  char inputMeshList[1000];
  unsigned int initNm=10000;
  unsigned int nPass = 1;
  unsigned int sparse_model = 1;
  float z = 0.5;
  float R = 0.85;
  unsigned int estTran = 1;
  unsigned int estModel = 1;
  char modelName[1000]="GMM.txt";
  unsigned int maxIter = 100; //changed from 100
  unsigned int findVirtual = 1;
  char virtualGenName[1000]="virt";
  char transformGenName[1000]="tran";
  int c = 0;
  char algList[1000]="algList.txt";
  char virList[1000]="virList.txt";
  char algGenName[1000]="alg";
  char virCorr[1000]="virCorr.txt";
  char logFile[1000]="logfile.txt";
  char convFileName[1000]="entropy.txt";

  unsigned int estAlpha = 0;
  while ((c = getopt (argc, argv, "n:i:N:c:s:z:t:m:r:g:v:T:M:a:A:V:C:L:p:x:R:")) != -1) 
   switch (c)
      {
      case 'n':
	nb_train = atoi(optarg);
	++rOpts;
	break;
      case 'i':
	sprintf(inputMeshList,"%s",optarg);
	++rOpts;
	break;
      case 'N':
	initNm = atoi(optarg);
       	break;
      case 'c':
	nPass = atoi(optarg);
	break;
      case 's':
	sparse_model = atoi(optarg);
	break;
      case 'z':
	z = atof(optarg);
	break;
      case 't':
	estTran = atoi(optarg);
	break;
      case 'm':
	estModel = atoi(optarg);
	break;
      case 'r':
	maxIter = atoi(optarg);
	break;
      case 'g':
	sprintf(virtualGenName,"%s",optarg);
	break;
      case 'x':
	sprintf(convFileName,"%s",optarg);
	break;
      case 'v':
	findVirtual = atoi(optarg);
	break;
      case 'T':
	sprintf(transformGenName,"%s", optarg);
	break;
      case 'M':
	sprintf(modelName,"%s", optarg);
	++rOpts;
	break;
      case 'a':
	sprintf(algGenName,"%s", optarg);
	break;
      case 'A':
	sprintf(algList,"%s", optarg);
	break;
      case 'V':
	sprintf(virList,"%s", optarg);
	break;
      case 'C':
	sprintf(virCorr,"%s", optarg);
	break;
      case 'L':
	sprintf(logFile,"%s", optarg);
	break;
      case 'p':
	estAlpha = atoi(optarg);
	break;
      case 'R':
	R = atof(optarg);
	break;
      case '?':
	fprintf (stderr, "Unknown option %s.\n", optarg);
	return 1;
      default:
	abort ();
      }
  
  if (rOpts != 3 )
    {
      print_help();
      return 1;
    }

  std::ofstream logfile;
  logfile.open(logFile);
  time_t rawtime;
  struct tm * timeinfo;

  time ( &rawtime );
  timeinfo = localtime ( &rawtime ); 
  logfile << asctime(timeinfo) << std::endl;
  for (int i=0; i< argc; i++)
    {
      logfile << argv[i] << " ";
    }
  // First we read meshes

  std::fstream meshNames;
  meshNames.open(inputMeshList);

  std::vector<char *> meshNamesVector(nb_train);
  for ( unsigned int k = 0; k < nb_train; k++ )
    {
      meshNamesVector[k] = new char[1000];
    }
  std::cout << "point sets read .. " << std::endl;
  //Next, we make pointset
  //Something to keep the point sets
  PointSetVectorType pointSetVector( nb_train );
  PointSetType::Pointer Datapoints =  PointSetType::New(); // store input data points

  int INDEX = 0; // this is to keep the number of total points
  for ( unsigned int k = 0; k < nb_train; k++ )
    {
      std::fstream myfile;
      meshNames >> meshNamesVector[k];
      myfile.open(meshNamesVector[k]);
      int index = 0;
      pointSetVector[k] = PointSetType::New();
      while( !myfile.eof() )
	{
	  std::string str;
	  std::getline(myfile,str);
	  if (!str.empty())
	    {
	      std::istringstream tmp(str);
	      VNLPointType vnlp;
	      tmp >> vnlp;
	      //std::cout << "vnp: " << vnlp << std::endl;
	      SetVNLPoint( pointSetVector[k],index++,&vnlp); 
	    }
	}
      // Store the data points in a PointSet vector
      Datapoints = pointSetVector[k];
      std::cout << "No. of data points =  " << Datapoints->GetNumberOfPoints() << std::endl;
      myfile.close();
      INDEX += index;
    }

  meshNames.close();


  if (estModel)
    {
      if ( initNm  > INDEX )
	initNm = INDEX - nb_train;
 
      std::cout << "initNm: " << initNm << std::endl;
      logfile << "initNm: " << initNm << std::endl;
    }
  

  //now I need to initilize the parameter sets including:
  // Transformaions
  AffineMatrixVectorType affineMatrices;
  TranslationVectorType translationVectors;
  for ( int k = 0; k < nb_train; k++)
    {
      AffineMatrixType rot;
      rot.set_identity();
      affineMatrices.push_back(rot);
      TranslationType t;
      t.fill(0.0);
      translationVectors.push_back(t);
    }
  // Posteriors(E's)
  PosteriorsMatrixVectorType posteriorsVector;
  

  double alpha = 0.0;
  double sigmaV[GMMPointDimension];
  // GMM Means
  VNLPointVectorType  modelPointsVector;
  // Priors
  PriorsVectorType priorsVector;
  PriorsVectorType uncPriorsVector;

  if ( estModel )
    {

      for ( int k = 0; k < nb_train; k++)
	{
	  int Ns = pointSetVector[k]->GetNumberOfPoints();
	  PosteriorsMatrixType post(Ns, initNm);
	  post.fill(1.0/initNm);
	  posteriorsVector.push_back(post);
	}

      PointSetType::Pointer initialMeanPointSet = PointSetType::New();
      InitializeGmmMeans(&pointSetVector, initialMeanPointSet, initNm, nPass);

      for (int j = 0; j < initNm; j++)
	{
	  double p = 1.0/initNm;
	  priorsVector.push_back(p);
	  uncPriorsVector.push_back(p);
	}
  
      PointsIteratorType it = initialMeanPointSet->GetPoints()->Begin();
      while (it != initialMeanPointSet->GetPoints()->End())
	{
	  
	  VNLPointType vnlp;
	  GetVNLPoint(initialMeanPointSet, it.Index() , &vnlp);
	  modelPointsVector.push_back(vnlp);
	  ++it;
	}
  
      UpdateSigma(&pointSetVector, &modelPointsVector, &affineMatrices, 
      		  &translationVectors, &posteriorsVector, sigmaV);
    }
  else 
    { // we read the model from a file
      std::ifstream myfile;
      myfile.open(modelName);
      int index = 0;
      while( !myfile.eof() )
	{
	  if ( index == 0  )
	    {
	      for(int i = 0; i<GMMPointDimension; i++) //this
	      	myfile >> sigmaV[i];
	    }
	  else
	    {
	      VNLPointType vnlp;
	      myfile >> vnlp;
	      modelPointsVector.push_back(vnlp);
	      double pi;
	      myfile >> pi;
	      priorsVector.push_back(pi);
	    }
	  ++index;
	}

      myfile.close();  
      initNm = index -1;
      for ( int k = 0; k < nb_train; k++)
	{
	  int Ns = pointSetVector[k]->GetNumberOfPoints();
	  PosteriorsMatrixType post(Ns, initNm);
	  post.fill(1.0/initNm);
	  posteriorsVector.push_back(post);
	}
    }

  if (!estAlpha)
    {
      if ( !sparse_model )
	z = 0.0;
      //std::cout << "z: " << z << std::endl;
      logfile << "z: " << z << std::endl;
      alpha = (1.0 - z)*1.0 + z*(1.0 - 1.0/initNm);
      
    }
  else
    {
      //z = .2;
      alpha = (1.0 - z)*1.0 + z*(1.0 - 1.0/initNm);
    }

  double dSigma[GMMPointDimension];
  double sigmaOld[GMMPointDimension];
  for(int i=0; i<GMMPointDimension; i++)
    { sigmaOld[i] =sigmaV[i]; 
      dSigma[i]= 1e32;}

  UpdateSigma(&pointSetVector, &modelPointsVector, &affineMatrices, 
        &translationVectors, &posteriorsVector, sigmaV);

  // for(int i=0; i<GMMPointDimension; i++)
  // std::cout << sigmaV[i] << std::endl;


  for( int i=0; i < maxIter; i++ )
    {

      std::cout <<"iteration: " << i+1 << std::endl;
      logfile <<"iteration: " << i+1 << std::endl;
      //--------------------------------------E-Step-------------------------------------
      UpdatePosteriorsVector(&pointSetVector, &modelPointsVector, &affineMatrices, 
      			     &translationVectors, &posteriorsVector, &priorsVector, sigmaV);
      //--------------------------------------M-Step-------------------------------------

      if (estTran)
	{
	  UpdateRigidTransformations(&pointSetVector, &modelPointsVector, &affineMatrices, 
	  			     &translationVectors, &posteriorsVector);

	}
      if (estModel)
	{
	  UpdateModelPointsVector(&pointSetVector, &modelPointsVector, &affineMatrices, 
				  &translationVectors, &posteriorsVector);
	}
      if ( sparse_model )
	{
	  std::cout << "updating sigma .." << std::endl;
	  UpdateSigma(&pointSetVector, &modelPointsVector, &affineMatrices, 
		      &translationVectors, &posteriorsVector, sigmaV);
	}
      else
	{
	  for(int i=0; i<GMMPointDimension; i++)
	    sigmaV[i] *= R;
	}

      UpdateMixtureCoeffs(&posteriorsVector, &priorsVector);
      /*    if ( estModel )
	{
	  //now updating model and priors ...
	  UpdateUnconstrainedPriorsVector(&posteriorsVector, &uncPriorsVector, &alpha);
	  UpdateConstrainedPriorsVector(&uncPriorsVector, &priorsVector);
	  PruneModel(&modelPointsVector, &priorsVector, &uncPriorsVector,&posteriorsVector);

	  std::cout <<"   total sparsity: " << (1.0 - (double(priorsVector.end() - priorsVector.begin())/(double)initNm))*100.0 << std::endl;
	  std::cout <<"   remaining live points: " << double(priorsVector.end() - priorsVector.begin()) << std::endl;
	  logfile <<"     total sparsity: " << (1.0 - (double(priorsVector.end() - priorsVector.begin())/(double)initNm))*100.0 << std::endl;
	  logfile <<"     remaining live points: " << double(priorsVector.end() - priorsVector.begin()) << std::endl;
    	}//estModel
*/
      //if (estModel)
      //	{
      std::cout << "sigma: ";
      for (int i1=0; i1<GMMPointDimension; i1++)
	{
	  std::cout <<  sigmaV[i1] << " ";
	  logfile << sigmaV[i1] << " ";
	}
      std::cout << std::endl;
      logfile << std::endl;
      //	}

      if (sparse_model)
	{
	  for(int i=0; i<GMMPointDimension; i++)
	    dSigma[i] = sigmaOld[i] - sigmaV[i];
	  for(int i=0; i<GMMPointDimension; i++)
	    sigmaOld[i] =sigmaV[i];
	  double d=0.0;
	  for(int i=0; i<GMMPointDimension; i++)
	    d += dSigma[i]*dSigma[i];
	  // if ( vcl_sqrt(d) < 1e-4 )
	  //	break;
	}
     }//EM
  
  if( estModel )
    {
      PointSetType::Pointer sparseModelPointSet =  PointSetType::New();
      VNLPointVectorType::iterator modelPointsVectorIt;
      modelPointsVectorIt = modelPointsVector.begin();
      while( modelPointsVectorIt != modelPointsVector.end())
	{
	  int j =  modelPointsVectorIt - modelPointsVector.begin();
	  PointType p;
	  for ( int k = 0; k<Dimension; k++)
	    p[k] = (*modelPointsVectorIt)[k];
	  sparseModelPointSet->SetPoint(j,p);
	  if (GMMPointDimension > Dimension )
	    {
	      VectorType v;
	      for ( int k = Dimension; k<GMMPointDimension; k++)
		v[ k - Dimension ] = (*modelPointsVectorIt)[k];
	      sparseModelPointSet->SetPointData(j,v);
	    }
	  ++modelPointsVectorIt;
	}

      std::cout << "writing gmm model ..." << std::endl;
      logfile << "writing gmm model ..." << std::endl;
      WriteModel(sparseModelPointSet, &priorsVector, modelName, sigmaV);
    }

  if (estTran)
    WriteTransformations(&affineMatrices, &translationVectors, &meshNamesVector, transformGenName );

  if (findVirtual)
    {
      std::cout << "writing correspondance ..." << std::endl;
      logfile << "writing correspondance ..." << std::endl;
      
      std::ofstream filelist;
      filelist.open(virList);

      std::vector< std::ofstream *> filesVector;
      
      int N = nb_train;
      for(int i=0; i< N; i++)
	{
	  std::string str= basename(meshNamesVector[i]);
	  unsigned found = str.find_last_of(".");
	  char bname[1000];
	  sprintf(bname,"%s",(str.substr(0,found)).c_str());
	  char filename[1000];
	  char currdir[1000];
      
	  char *unusd=getcwd(currdir,1000);
	  sprintf(filename,"%s/%s_%s%s",currdir,bname,virtualGenName,".txt");
 	  //std::cout << filename << std::endl;
	  filelist << filename << std::endl;
	  std::ofstream * myfile= new std::ofstream;
	  myfile->open(filename);
	  filesVector.push_back(myfile);
	}
      filelist.close();

      WriteCorrespondance(&pointSetVector, &modelPointsVector, &affineMatrices, 
			  &translationVectors, &posteriorsVector, &filesVector, virCorr);
    }

  if ( estTran )
    {
      std::ofstream filelist;
      filelist.open(algList);

      std::vector< std::ofstream *> filesVector;
      
      int N = nb_train;
      for(int i=0; i< N; i++)
	{
	  std::string str= basename(meshNamesVector[i]);
	  unsigned found = str.find_last_of(".");
	  char bname[1000];
	  sprintf(bname,"%s",(str.substr(0,found)).c_str());
	  char filename[1000];
	  char currdir[1000];
      
	  char *unusd=getcwd(currdir,1000);
	  sprintf(filename,"%s/%s_%s%s",currdir,bname,algGenName,".txt");
 	  //sprintf(filename,"%s_%s%s",str.substr(0,found).c_str(),algGenName,".txt");
	  //std::cout << filename << std::endl;
	  filelist << filename << std::endl;
	  std::ofstream * myfile= new std::ofstream;
	  myfile->open(filename);
	  filesVector.push_back(myfile);
	}
      filelist.close();

      TransformShapes(&pointSetVector, &affineMatrices,&translationVectors, &filesVector);
      
    }

  for ( unsigned int k = 0; k < nb_train; k++ )
    {
      delete meshNamesVector[k];
    }
  
  std::cout << "finished." << std::endl;
  logfile << "finished." << std::endl;
  
  logfile.close();


  return EXIT_SUCCESS;
}
